from marquetry import Function
from marquetry import functions


class Linear(Function):
    """Perform a linear transformation on an input array.

        This class defines a function that performs a linear transformation on an input array using
        the provided weight matrix and optional bias.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def forward(self, x, w, b):
        y = x.dot(w)
        if b is not None:
            y += b
        return y

    def backward(self, inputs, grad_y):
        x, w, b = inputs
        grad_y = grad_y[0]

        grad_b = None if b is None else functions.sum_to(grad_y, b.shape)

        grad_x = functions.matmul(grad_y, w.T)
        grad_w = functions.matmul(x.T, grad_y)

        return grad_x, grad_w, grad_b


def linear(x, w, b=None):
    """Perform a linear transformation on an input array or container.


        This class defines a function that performs a linear transformation on an input array using
        the provided weight matrix and optional bias.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input array to be linearly transformed.
            w (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The weight matrix for the linear transformation.
            b (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray` or None):
                The bias term. If None, no bias is added.

        Returns:
            :class:`marquetry.Container`: The result of the linear transformation on
                the input array using the provided weight matrix and bias (if any).

        Examples:
            >>> x = np.arange(0, 20).reshape(4, 5)
            >>> w = np.arange(0, 30).reshape(5, 6)
            >>> b = np.arange(0, 6)
            >>> linear(x, w, b)
            container([[ 180  191  202  213  224  235]
                       [ 480  516  552  588  624  660]
                       [ 780  841  902  963 1024 1085]
                       [1080 1166 1252 1338 1424 1510]])
            >>> linear(x, w, b).shape
            (4, 6)
    """
    return Linear()(x, w, b)



import os

import numpy as np
import pandas as pd

from marquetry import dataset, preprocesses
from marquetry.transformers import compose, astypes
from marquetry.utils import get_file


class Titanic(dataset.Dataset):
    """Get the Titanic dataset.

        Data obtained from http://hbiostat.org/data courtesy of the Vanderbilt University Department of Biostatistics.

        The sinking of the Titanic is one of the most infamous shipwrecks in history.

        On April 15, 1912, during her maiden voyage, the widely considered “unsinkable” RMS Titanic sank after colliding
        with an iceberg. Unfortunately, there weren't enough lifeboats for everyone onboard,
        resulting in the death of 1502 out of 2224 passengers and crew.
        While there was some element of luck involved in surviving,
        it seems some groups of people were more likely to survive than others.
        In this challenge, we ask you to build a predictive model that answers the question:
        “what sorts of people were more likely to survive?”
        using passenger data (ie name, age, gender, socio-economic class, etc).
        (From kaggle competition description.)

        Args:
            train (bool): If True, loads the training dataset.
                Default is True. (When you want to get test data, you should use :meth:`test_data`.)
            train_rate (float, optional): The fraction of data to use for training when `train` is True.
                Should be in the range (0.0, 1.0). Default is 0.8.
            label_columns (list): List of column names to be treated as labels.
                Default is None.
            drop_columns (list): List of column names to be dropped from the dataset.
                Default is None.
            pre_process (bool): If True, preprocesses the dataset by encoding, imputing, and normalizing columns.
                Default is True.
            down_float_bit (bool): If True, downcasts float data to float32 for memory efficiency.
                Default is True.

        Examples:
            >>> from marquetry.datasets import Titanic
            >>> train_dataset = Titanic(train=True, train_rate=0.8, label_columns=["sex"], pre_process=True)
            >>> test_dataset = train_dataset.test_data()
            >>> print(len(train_dataset), len(test_dataset))
            1047 262
            >>> print(train_dataset[0])
            (array([-0.17157276,  0.4409694, -0.44408074, ...,  0., 0.,  0.], dtype=float32), 0)
            >>> print(train_dataset.source_columns)
            ['age', 'sibsp', 'parch', 'fare', 'sex', 'pclass_2', 'pclass_3', ..., 'cabin_C82', 'ticket_unknown']
            >>> print(train_dataset.target_columns)
            survived
            >>> print(train_dataset.unique_numbers)
            {'pclass': 3, 'survived': 'N/A', ..., 'fare': 'N/A', 'cabin': 157, 'embarked': 3}

        Note:
            The Titanic dataset is commonly used for binary classification tasks.
            It contains information about passengers on the Titanic, including their features and whether they survived.

        Caution:
            When you want to get test data, we suggest to use :meth:`test_data` instead of the ``train=False``
            due to Titanic dataset needs to be set the same setting as the train data when you get test data.

            If the setting is changed, the test data can't get correctly.
            From :meth:`test_data`, you can get the test dataset correctly.

    """

    def __init__(self, train=True, train_rate=0.8, label_columns: list = None,
                 drop_columns: list = None, pre_process: bool = True, down_float_bit: bool = True):

        if not 0.0 <= train_rate <= 1.0:
            raise ValueError("train_rate should in the range of (0.0, 1.0), but got {}".format(train_rate))

        self._category_columns = ["pclass", "sex", "cabin", "embarked", "name", "ticket"]
        self._numerical_columns = ["age", "sibsp", "parch", "fare"]
        self._target_column = "survived"

        label_columns = label_columns if label_columns is not None else []
        if len(label_columns) == 0:
            one_hot_columns = self._category_columns
        else:
            one_hot_columns = [column for column in self._category_columns if column not in label_columns]

        self._label_columns = label_columns
        self._one_hot_columns = one_hot_columns

        self.train_rate = train_rate

        self.pre_process = pre_process
        self.drop_columns = drop_columns if drop_columns is not None else []

        self._source_columns = None
        self._unique_dict = None

        self._down_float_bit = down_float_bit
        if down_float_bit and pre_process:
            transform = compose.Compose([astypes.AsType(dtype=np.float32)])
        else:
            transform = None

        super().__init__(train, transform, None)

    def _set_data(self, **kwargs):
        url = "https://biostat.app.vumc.org/wiki/pub/Main/DataSets/titanic3.csv"

        data_path = get_file(url)
        data = self._load_data(data_path)

        source = data.drop("survived", axis=1)
        self._source_columns = list(source.columns)

        self.source = source.to_numpy()
        self.target = data.loc[:, "survived"].astype(int).to_numpy()

    def _load_data(self, file_path):
        titanic_df = pd.read_csv(file_path)
        titanic_df = self._remove_leak_column(titanic_df, file_path)

        train_last_index = int(len(titanic_df) * self.train_rate)
        if self.train:
            titanic_df = titanic_df.iloc[:train_last_index, :]
        else:
            titanic_df = titanic_df.iloc[train_last_index:, :]

        if self.pre_process:
            preprocess = preprocesses.ToEncodeData(
                target_column="survived", category_columns=self._category_columns,
                numeric_columns=self._numerical_columns, name="titanic_dataset", is_train=self.train,
                imputation_category_method="mode", imputation_numeric_method="median",
                label_encode_columns=self._label_columns, one_hot_encode_columns=self._one_hot_columns,
                normalize_method="normalize"
            )

            titanic_df, self._unique_dict = preprocess(titanic_df)

        for drop_column in self.drop_columns:
            drop_regex = titanic_df.filter(regex=drop_column)
            if drop_regex.shape[1] != 0:
                titanic_df = titanic_df.drop(drop_regex.columns, axis=1)

        return titanic_df

    @staticmethod
    def _remove_leak_column(data: pd.DataFrame, file_path):
        leak_columns = ("body", "home.dest", "boat")

        original_columns = data.columns
        for leak_column in leak_columns:
            if leak_column in data:
                data = data.drop(leak_column, axis=1)

        if any(column in original_columns for column in leak_columns):
            data = data.sample(frac=1, random_state=2023)
            data.reset_index(drop=True, inplace=True)
            data.to_csv(file_path, index=False)

        return data

    def test_data(self):
        test_dataset = Titanic(train=False, train_rate=self.train_rate, label_columns=self._label_columns,
                               drop_columns=self.drop_columns, pre_process=self.pre_process,
                               down_float_bit=self._down_float_bit)
        test_dataset._unique_dict = "Test data can't calculate the unique nums."

        return test_dataset

    @property
    def source_columns(self):
        return self._source_columns

    @property
    def target_columns(self):
        return self._target_column

    @property
    def unique_numbers(self):
        return self._unique_dict



import marquetry
from marquetry import functions
from marquetry import Layer


class BiLSTM(Layer):
    """Bidirectional Long Short-Term Memory (BiLSTM) layer for sequence modeling.

        The BiLSTM layer is a variation of the standard LSTM layer
        that processes input data in both forward and reverse directions.
        This allows the network to capture information from both past and future context,
        resulting in richer representations.
        About LSTM, please see :class:`marquetry.layers.LSTM`.

        Args:
            hidden_size (int): The size of the hidden state in each LSTM layer.
            in_size (int or None): The size of the input data.

        Caution:
            in_size:
                This is automatically determined from the input data shape
                and does not need to be specified except a special use case.

        Attributes:
            forward_lstm (:class:`marquetry.layers.LSTM`): Forward LSTM layer.
            reverse_lstm (:class:`marquetry.layers.LSTM`): Reverse LSTM layer.

        Examples:
            >>> dataset = marquetry.datasets.SinCurve()
            >>> dataloader = marquetry.dataloaders.SeqDataLoader(dataset, batch_size=32)
            >>> model = (BiLSTM(128), marquetry.layers.Linear(1))
            >>> loss = 0
            >>> for x, t in dataloader:
            >>>     for layer in model:
            >>>         x = layer(x)
            >>>     loss += functions.mean_squared_error(x, t)
            >>> loss
            container(36.87500179632384)

    """

    def __init__(self, hidden_size, in_size=None):
        super().__init__()
        self.forward_lstm = marquetry.layers.LSTM(hidden_size, in_size=in_size)
        self.reverse_lstm = marquetry.layers.LSTM(hidden_size, in_size=in_size)

    def reset_state(self):
        self.forward_lstm.reset_state()
        self.reverse_lstm.reset_state()

    def set_state(self, h, c=None):
        """Set the hidden state and cell state to a custom value.

            Args:
                h (marquetry.Container): The custom hidden state.
                c (marquetry.Container or None): The custom cell state.

            Caution:
                Almost general use case, the cell state should NOT set custom value
                because cell state in LSTM is used only internal information connection,
                and it should be managed automatically.
                If you don't have any special reason, you should set only hidden state.

        """

        self.forward_lstm.set_state(h, c)
        self.reverse_lstm.set_state(h, c)

    def forward(self, x):
        out1 = self.forward_lstm(x)
        out2 = self.reverse_lstm(x[:, ::-1])
        out2 = out2[:, ::-1]

        output = functions.concat((out1, out2), axis=-1)

        return output



import gzip

import numpy as np

from marquetry import dataset
from marquetry.utils import get_file
from marquetry.transformers import Flatten


class MNIST(dataset.Dataset):
    """Get the MNIST dataset.

        This dataset is sourced from http://yann.lecun.com/exdb/mnist/.

        The MNIST database of handwritten digits has a training set of 60,000 examples,
        and a test set of 10,000 examples.

        The Attributes and Args is following the :class:`marquetry.dataset.Dataset`, please check it.

        Examples:
            >>> data = MNIST(transform=Flatten())
            >>> data.source.shape
            (60000, 1, 28, 28)
            If the data extruct directly by ``.sorce``, the tranform doesn't apply.
            >>> source, target = data[0]
            >>> source.shape
            (784,)
    """
    def __init__(self, train=True, transform=None, target_transform=None):
        super().__init__(train, transform, target_transform)

    def _set_data(self):
        url = "http://yann.lecun.com/exdb/mnist/"

        train_files = {
            "source": "train-images-idx3-ubyte.gz",
            "target": "train-labels-idx1-ubyte.gz"
        }
        test_files = {
            "source": "t10k-images-idx3-ubyte.gz",
            "target": "t10k-labels-idx1-ubyte.gz"
        }

        files = train_files if self.train else test_files
        source_path = get_file(url + files["source"])
        target_path = get_file(url + files["target"])

        self.source = self._load_source(source_path)
        self.target = self._load_target(target_path)

    @staticmethod
    def _load_source(file_path):
        with gzip.open(file_path, "rb") as f:
            data = np.frombuffer(f.read(), np.uint8, offset=16)

        source = data.reshape((-1, 1, 28, 28))

        return source

    @staticmethod
    def _load_target(file_path):
        with gzip.open(file_path, "rb") as f:
            target = np.frombuffer(f.read(), np.uint8, offset=8)

        return target

    @property
    def labels(self):
        return {0: "0", 1: "1", 2: "2", 3: "3", 4: "4", 5: "5", 6: "6", 7: "7", 8: "8", 9: "9"}




import numpy as np

import marquetry
from marquetry import Function


class Absolute(Function):
    """Calculate the absolute value of an input.

        This class computes the absolute value of the input.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def forward(self, x):
        y = np.abs(x)

        return y

    def backward(self, x, grad_y):
        mask = (x[0] >= 0).astype("f")
        mask -= .5
        mask *= 2.

        grad_y = grad_y[0]
        if mask.shape != grad_y.shape:
            marquetry.functions.broadcast_to(grad_y, x[0].shape)

        grad_x = grad_y * mask

        return grad_x


def absolute(x):
    """Calculate the absolute value of an input.

        This function computes the absolute value of the input.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input values.

        Returns:
            :class:`marquetry.Container`: The absolute value of the input.

        Examples:
            >>> x = np.array([[-2, 4, 0], [-1, 2, -3]])
            >>> x
            array([[-2, 4, 0],
                   [-1, 2, -3]])
            >>> absolute(x)
            container([[2 4 6]
                       [1 2 3]])
    """

    return Absolute()(x)



from marquetry import cuda_backend
from marquetry import Function


class FScore(Function):
    """Compute the F-score based on true labels and predicted values using a specified threshold.

        This class defines a function that computes the F-score based on the true binary labels and predicted values.

        The F-score is a way of combining the precision and recall of the model,
        and it is defined as the harmonic mean of the model’s precision and recall.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.

    """

    def __init__(self, threshold, dtype="float64"):
        if not 0. <= threshold <= 1.:
            raise ValueError("threshold should be in (0.0, 1.0), but got {}".format(threshold))

        self.threshold = threshold
        self.dtype = dtype

    def forward(self, y, t):
        xp = cuda_backend.get_array_module(y)

        assert len(xp.unique(t)) <= 2

        pred = xp.asarray((y >= self.threshold), dtype="f").reshape(t.shape)

        true_positive_num = pred[t == 1].sum()
        pred_positive_num = pred.sum()
        target_positive_num = xp.asarray((t == 1), dtype="f").sum()

        precision_value, recall_value = _precision_recall_validator(true_positive_num, pred_positive_num,
                                                                    target_positive_num, xp=xp)

        self.retain_inputs(())
        if precision_value == 0. and recall_value == 0.:
            return xp.asarray(0.0, dtype=self.dtype)
        else:
            f_score_value = 2 * precision_value * recall_value / (precision_value + recall_value)
            return xp.asarray(f_score_value.data, dtype=self.dtype)


def f_score(y, t, threshold=0.7, dtype="float64"):
    """Compute the F-score based on true labels and predicted values using a specified threshold.

        This function defines that computes the F-score based on the true binary labels and predicted values.

        The F-score is a way of combining the precision and recall of the model,
        and it is defined as the harmonic mean of the model’s precision and recall.

        F-score is calculated as:
            :math:`F-score = (2 * precision * recall) / (precision + recall)`

        More species, please see :meth:`marquetry.functions.precision` and :meth:`marquetry.functions.recall`.

        Args:
            y (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted values.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true binary labels.
            threshold (float): The threshold value used to determine binary predictions.
                Defaults to 0.7.
            dtype (str or numpy.dtype):
                The return value's dtype, Default is "float64" that means 64-bit float value.

        Returns:
            :class:`marquetry.Container`: The F-score based on the true binary labels
                and predicted values using the specified threshold.
    """

    return FScore(threshold, dtype)(y, t)


class MultiFScore(Function):
    """Compute the F-score for a specific target class based on true labels and predicted values.

        This class defines a function that computes the F-score for a specific target class based on
        the true labels and predicted values.

        The F-score is a way of combining the precision and recall of the model,
        and it is defined as the harmonic mean of the model’s precision and recall.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.

    """

    def __init__(self, target_class, dtype="float64"):
        self.target_class = target_class
        self.dtype = dtype

    def forward(self, y, t):
        xp = cuda_backend.get_array_module(y)

        assert len(xp.unique(t)) > 2

        if y.size != t.size:
            pred = xp.argmax(y, axis=1).reshape(t.shape)
        else:
            pred = y

        pred_positive = xp.asarray(pred == self.target_class, dtype="f")
        true_map = xp.asarray(pred == t, dtype="f")

        true_positive_num = xp.asarray(pred_positive * true_map, dtype="f").sum()
        pred_positive_num = pred_positive.sum()
        target_positive_num = xp.asarray(t == self.target_class, dtype="f").sum()

        precision_value, recall_value = _precision_recall_validator(true_positive_num, pred_positive_num,
                                                                    target_positive_num, xp=xp)

        self.retain_inputs(())
        if precision_value == 0. and recall_value == 0.:
            return xp.asarray(0.0, dtype=self.dtype)
        else:
            f_score_value = 2 * precision_value * recall_value / (precision_value + recall_value)
            return xp.asarray(f_score_value.data, dtype=self.dtype)


def multi_f_score(y, t, target_class=1, dtype="float64"):
    """Compute the F-score for a specific target class based on true labels and predicted values.

        This function defines that computes the F-score for a specific target class
        based on the true labels and predicted values.
        The F-score is a way of combining the precision and recall of the model,
        and it is defined as the harmonic mean of the model’s precision and recall.

        F-score is calculated as:
            :math:`F-score = (2 * precision * recall) / (precision + recall)`

        More species,
        please see :meth:`marquetry.functions.multi_precision` and :meth:`marquetry.functions.multi_recall`.

        Args:
            y (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted values.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true labels.
            target_class (int):
                The class index to be treated as the positive class when calculating the F-score.
                Defaults to 1.
            dtype (str or numpy.dtype):
                The return value's dtype, Default is "float64" that means 64-bit float value.

        Returns:
            :class:`marquetry.Container`: The F-score for the specified target class
                based on the true labels and predicted values.
    """
    return MultiFScore(target_class, dtype)(y, t)


def _precision_recall_validator(true_positive_num, pred_positive_num, target_positive_num, xp):
    """Validate pred_positive_num and target_positive_num to avoid Zero-Division and calc the precision and recall.

        Args:
            true_positive_num (int): The true positive num means a model predicts a ``True``,
                and it is actuary ``True`` counts.
            pred_positive_num (int): The true pred num means a model predicts a ``True`` counts.
            target_positive_num (int): The real true num means a target label's ``True`` counts.

        Returns:
            :class:`cupy.ndarray` or :class:`numpy.ndarray`: The precision and recall value.
    """
    if pred_positive_num == 0 and target_positive_num == 0:
        precision_value = xp.asarray(0.0)
        recall_value = xp.asarray(0.0)
    elif pred_positive_num == 0:
        precision_value = xp.asarray(0.0)
        recall_value = xp.asarray(true_positive_num / target_positive_num)
    elif target_positive_num == 0:
        precision_value = xp.asarray(true_positive_num / pred_positive_num)
        recall_value = xp.asarray(0.0)
    else:
        precision_value = xp.asarray(true_positive_num / pred_positive_num)
        recall_value = xp.asarray(true_positive_num / target_positive_num)

    return precision_value, recall_value



from marquetry import Function


class MeanSquaredError(Function):
    """Calculate the Mean Squared Error (MSE) between two input arrays.

        This class computes the Mean Squared Error (MSE) between two input arrays `x0` and `x1`.
        The MSE is a measure of the average squared difference between corresponding elements  of `x0` and `x1`.
        It is often used as a loss function in regression problems.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def forward(self, x0, x1):
        diff = x0 - x1
        y = (diff ** 2).sum() / diff.dtype.type(diff.size)

        return y

    def backward(self, inputs, grad_y):
        x0, x1 = inputs
        diff = x0 - x1
        grad_x0 = 2. * diff / diff.size * grad_y[0]
        grad_x1 = -grad_x0 * grad_y[0]

        return grad_x0, grad_x1


def mean_squared_error(x0, x1):
    """Calculate the Mean Squared Error (MSE) between two input arrays.

        This function computes the Mean Squared Error (MSE) between two input arrays `x0` and `x1`.
        The MSE is a measure of the average squared difference between corresponding elements of `x0` and `x1`.
        It is often used as a loss function in regression problems.

        Parameters:
            x0 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The first input array.
            x1 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The second input array.

        Returns:
            :class:`marquetry.Container`: The Mean Squared Error (MSE).
    """

    return MeanSquaredError()(x0, x1)



from marquetry import Function


class Reshape(Function):
    """Reshape an input array to the specified shape.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def __init__(self, shape):
        self.shape = shape

        self.x_shape = None

    def forward(self, x):
        self.x_shape = x.shape
        y = x.reshape(self.shape)

        self.retain_inputs(())
        return y

    def backward(self, x, grad_y):
        return reshape(grad_y[0], self.x_shape)


def reshape(x, shape):
    """This function reshapes an input array to the specified shape.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input array to be reshaped.
            shape (tuple or int): The target shape to which the input array should be reshaped.

        Returns:
            A :class:`marquetry.Container` object which is the result of reshaping
            the input array to the specified shape.

        Examples:
            >>> x = np.arange(0, 3).reshape(1, 3)
            array([[0, 1, 2]])
            >>> reshape(x, shape=3)
            container([0 1 2])
    """

    return Reshape(shape)(x)



from marquetry import cuda_backend
from marquetry.optimizer import Optimizer


class RMSProp(Optimizer):
    """Root Mean Square Propagation(RMSProp) optimizer for updating model parameters during training.

        RMSProp is an optimization algorithm that adapts the learning rates for each parameter
        based on the past gradients.
        It helps prevent the learning rate from being too large, which can lead to divergence.

        The `decay` parameter controls the rate of decay for the moving average.

        Args:
            learning_rate (float): The learning rate for updating parameters.
                Default is 0.01.
            decay (float): The decay rate for the moving average of past gradients.
                Default is 0.99.
            eps (float): A small constant added to the denominator to prevent division by zero.
                Default is 1e-8.

        Tip:
            When you would like to optimize your model's parameter,
            please set the model to this using ``prepare`` method.

        Examples:
            >>> optimizer = RMSProp()
            >>> model = marquetry.models.MLP([128, 256, 64, 10])
            >>> optimizer.prepare(model)
            >>> optimizer.update()

    """

    def __init__(self, learning_rate=0.01, decay=0.99, eps=1e-8):
        super().__init__()
        self.lr = learning_rate
        self.decay = decay
        self.eps = eps

        self.histories = {}

    def _update_one(self, param):
        h_key = id(param)

        xp = cuda_backend.get_array_module(param.data)
        if h_key not in self.histories:
            self.histories[h_key] = xp.zeros_like(param.data)

        history = self.histories[h_key]
        grad = param.grad.data

        history *= self.decay
        history += (1 - self.decay) * grad ** 2

        param.data -= self.lr * grad / (xp.sqrt(history) + self.eps)



import numpy as np

import marquetry


# ===========================================================================
# Dataset base class
# ===========================================================================
class Dataset(object):
    """Dataset base class that provides data format function.

    All dataset implementations defined in :mod:`marquetry.datasets` inherit this class.


    The main feature of this class is providing data and target following user setting.
    If you specify the :attr:`train` is True, this provides train dataset and if False,
    provides test dataset.

    Args:
        train (bool): toggle if you want to load train_data or not.
        transform (:class:`marquetry.transformers.Compose` or class or None):
            transform means transform **data** that is used when data loaded from anything.
        target_transform(:class:`marquetry.transformers.Compose` or class or None)
            : transform means transform **target** that is used when data loaded from anything.

    Note:
        This object is usually used with DataLoader object.

    """
    def __init__(self, train=True, transform=None, target_transform=None):
        self.train = train
        self.transform = transform
        self.target_transform = target_transform

        if self.transform is None:
            self.transform = lambda x: x

        if self.target_transform is None:
            self.target_transform = lambda x: x

        self.source = None
        self.target = None

        self._set_data()

    def __getitem__(self, index):
        assert np.isscalar(index)
        if self.target is None:
            return self.transform(self.source[index]), None
        else:
            return self.transform(self.source[index]), self.target_transform(self.target[index])

    def __len__(self):
        return len(self.source)

    @property
    def source_shape(self):
        return self.source.shape

    def _set_data(self, *args):
        raise NotImplementedError()



from marquetry import Function


class MatMul(Function):
    """Calculate dot product between two inputs.

        This class calculates dot product between the two input matrices.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def forward(self, x0, x1):
        y = x0.dot(x1)

        return y

    def backward(self, xs, grad_y):
        x0, x1 = xs
        grad_y = grad_y[0]

        grad_x0 = matmul(grad_y, x1.T)
        grad_x1 = matmul(x0.T, grad_y)

        return grad_x0, grad_x1


def matmul(x0, x1):
    """Calculate dot product between two inputs.

        This function calculates dot product between the two input matrices.

        Args:
            x0 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The first input matrix.
            x1 (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The second input matrix.

        Returns:
            :class:`marquetry.Container`: The result of the dot product.

        Examples:
            >>> x0 = np.arange(0, 20).reshape(5, 4)
            >>> x0
            array([[ 0,  1,  2,  3],
                   [ 4,  5,  6,  7],
                   [ 8,  9, 10, 11],
                   [12, 13, 14, 15],
                   [16, 17, 18, 19]])
            >>> x1 = np.arange(0, 24).reshape(4, 6)
            >>> x1
            array([[ 0,  1,  2,  3,  4,  5],
                   [ 6,  7,  8,  9, 10, 11],
                   [12, 13, 14, 15, 16, 17],
                   [18, 19, 20, 21, 22, 23]])
            >>> matmul(x0, x1)
            container([[  84   90   96  102  108  114]
                       [ 228  250  272  294  316  338]
                       [ 372  410  448  486  524  562]
                       [ 516  570  624  678  732  786]
                       [ 660  730  800  870  940 1010]])
            >>> matmul(x0, x1).shape
            (5, 6)

    """

    return MatMul()(x0, x1)



def pair(x):
    """Get a pair of ints.

        This function takes an input and returns a pair of values.
        If the input is an integer, it creates a pair of identical values.
        If the input is already a pair of values (tuple of length 2), it returns the input as is.

        For other types of input, a ValueError is raised.

        Args:
            x (int or tuple): The input value to create a pair from.

        Returns:
            tuple: A pair of ints.
    """

    if isinstance(x, int):
        return x, x
    elif isinstance(x, tuple):
        assert len(x) == 2
        return x
    else:
        raise ValueError("pair can't use {}".format(x))



import numpy as np

from marquetry import cuda_backend
from marquetry import dataloaders


class SeqDataLoader(dataloaders.DataLoader):
    """A specialized data loader for sequential data.

        The `SeqDataLoader` class is a subclass of the `DataLoader` designed specifically for loading sequential data.
        It inherits most of its behavior from the parent `DataLoader` class
        but modifies the batch creation logic to maintain sequence order.

        Args:
            dataset (:class:`marquetry.dataset.Dataset`): The dataset to be loaded and batched.
            batch_size (int): The size of each batch
            cuda (bool): Whether to use GPU (CUDA) for data storage (if available).

        Examples:
            >>> seq_dataset = SequentialDataset()
            >>> seq_dataloader = SeqDataLoader(seq_dataset, batch_size=32, cuda=True)
            Iterating over batches of sequential data while maintaining order:
            >>> for batch_x, batch_t in seq_dataloader:
            Process the batch of input data (batch_x) and target data (batch_t).

        Caution:
            Generally, DataLoader object is used with ``for`` statement.
            If you have no special reason, you don't need ``reset`` or ``next`` method.
    """
    def __init__(self, dataset, batch_size, cuda=False):
        super().__init__(dataset=dataset, batch_size=batch_size, shuffle=False, cuda=cuda)

    def __next__(self):
        if self.iterations >= self.max_iters:
            self.reset()
            raise StopIteration

        jump = self.data_size // self.batch_size
        batch_index = [(i * jump + self.iterations) % self.data_size for i in range(self.batch_size)]

        batch = [self.dataset[i] for i in batch_index]

        xp = cuda_backend.cp if self.cuda else np
        x = xp.array([example[0] for example in batch])
        t = xp.array([example[1] for example in batch])

        self.iterations += 1

        return x, t



from marquetry import cuda_backend
from marquetry import Function
from marquetry import functions


class BroadcastTo(Function):
    """BroadcastTo function that transform the array to specify shape.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def __init__(self, shape):
        self.shape = shape

        self.x_shape = None

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        if x.shape == self.shape:
            return x

        self.x_shape = x.shape
        y = xp.broadcast_to(x, self.shape)

        self.retain_inputs(())
        return y

    def backward(self, x, grad_y):
        if self.x_shape is None:
            return grad_y[0]

        grad_x = functions.sum_to(grad_y[0], self.x_shape)

        return grad_x


def broadcast_to(x, shape):
    """BroadcastTo function that transform the array to specify shape.

        This function copies the array to the specifying shape.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                Input container that is :class:`marquetry.Container` array.
            shape (tuple): The goal shape using this function.

        Returns:
            marquetry.Container: Output container. A float array.

        Examples:

            >>> x = np.array([[1, -3]], 'f')
            >>> x
            array([[ 1., -3.]], dtype=float32)
            >>> broadcast_to(x, shape=(4, 2))
            container([[ 1. -3.]
                       [ 1. -3.]
                       [ 1. -3.]
                       [ 1. -3.]])

        """
    return BroadcastTo(shape)(x)



import marquetry
from marquetry import cuda_backend


def col2im_array(col, img_shape, kernel_size, stride, pad, to_matrix=True):
    """The util of col2im main process"""

    batch_size, channels, height, width = img_shape
    kernel_height, kernel_width = marquetry.utils.pair(kernel_size)
    stride_height, stride_width = marquetry.utils.pair(stride)
    padding_height, padding_width = marquetry.utils.pair(pad)

    out_height = marquetry.utils.get_convolution_outsize(height, kernel_height, stride_height, padding_height)
    out_width = marquetry.utils.get_convolution_outsize(width, kernel_width, stride_width, padding_width)

    if to_matrix:
        # Reconvert to 6-dims array from the easy used matrix to convolut.
        col = (col.reshape(batch_size, out_height, out_width, channels, kernel_height, kernel_width).
               transpose(0, 3, 4, 5, 1, 2))

    xp = cuda_backend.get_array_module(col)
    # prepare the clear image array which is 4-dims.
    img = xp.zeros(
        (
            batch_size,
            channels,
            height + 2 * padding_height + stride_height - 1,
            width + 2 * padding_width + stride_width - 1), dtype=col.dtype)

    for height_range in range(kernel_height):
        height_lim = height_range + stride_height * out_height

        for width_range in range(kernel_width):
            width_lim = width_range + stride_width * out_width

            img[:, :, height_range:height_lim:stride_height, width_range:width_lim:stride_width] += col[:, :, height_range, width_range, :, :]

    return img[:, :, padding_height:height + padding_height, padding_width:width + padding_width]


from marquetry.utils.array.col2im_array import col2im_array
from marquetry.utils.array.im2col_array import im2col_array

from marquetry.utils.gradients.numerical_grad import numerical_grad

from marquetry.utils.io.get_file import get_file

from marquetry.utils.math.backward_utils import max_backward_shape
from marquetry.utils.math.backward_utils import reshape_sum_backward
from marquetry.utils.math.log_sum_exp import log_sum_exp
from marquetry.utils.math.sum_to import sum_to
from marquetry.utils.math.truncate_data import floor
from marquetry.utils.math.truncate_data import ceil

from marquetry.utils.numeric.get_size import get_convolution_outsize
from marquetry.utils.numeric.get_size import get_deconvolution_outsize
from marquetry.utils.numeric.pair import pair

from marquetry.utils.tests.array_close import array_close
from marquetry.utils.tests.array_equal import array_equal
from marquetry.utils.tests.gradient_check import gradient_check

from marquetry.utils.tree_utils.impurity_criterion import impurity_criterion
from marquetry.utils.tree_utils.information_gain import information_gain
from marquetry.utils.tree_utils.split_branch import split_branch

from marquetry.utils.visualize.dot_graph import get_dot_graph
from marquetry.utils.visualize.dot_graph import plot_dot_graph



import numpy as np

from marquetry import dataset


class TrigonometricCurve(dataset.Dataset):
    """Get SinCurve dataset.

        Toy problem that is trigonometric number between 0 and 3π.
        The train dataset is sin curve with small noise, and the test dataset is cos curve.

        The Attributes and Args is following the :class:`marquetry.dataset.Dataset`, please check it.

        Examples:
            >>> data = TrigonometricCurve()
            >>> data.source.shape
            (4999, 1)
            >>> source, target = data[0]
            >>> source.shape
            (1,)
    """
    def _set_data(self):
        num_data = 5000
        dtype = np.float64

        x = np.linspace(0, 3 * np.pi, num_data)
        noise_range = (-0.05, 0.05)
        noise = np.random.uniform(noise_range[0], noise_range[1], size=x.shape)

        if self.train:
            y = np.sin(x) + noise
        else:
            y = np.cos(x)

        y = y.astype(dtype)
        self.target = y[1:][:, np.newaxis]
        self.source = y[:-1][:, np.newaxis]



from marquetry import cuda_backend
from marquetry import Function
from marquetry import functions
from marquetry import utils


class Conv2DGradW(Function):
    """Weight gradient calculation class for Convolution2D backward."""
    def __init__(self, conv2d_instance):
        w = conv2d_instance.inputs[1]
        kernel_height, kernel_width = w.shape[2:]
        self.kernel_size = (kernel_height, kernel_width)
        self.stride = conv2d_instance.stride
        self.pad = conv2d_instance.pad

    def forward(self, x, grad_y):
        xp = cuda_backend.get_array_module(x)

        col = utils.im2col_array(x, self.kernel_size, self.stride, self.pad, to_matrix=False)
        grad_w = xp.tensordot(grad_y, col, ((0, 2, 3), (0, 4, 5)))

        self.retain_outputs((0,))
        return grad_w

    def backward(self, inputs, grad_ys):
        x, grad_y = inputs
        grad_w, = self.output_data

        x_height, x_width = x.shape[2:]
        grad_x = functions.deconvolution_2d(
            grad_y, grad_w, stride=self.stride, pad=self.pad, out_size=(x_height, x_width))
        grad_grad_y = functions.convolution_2d(x, grad_w, stride=self.stride, pad=self.pad)

        return grad_x, grad_grad_y



import os
import urllib.request

from marquetry.configuration import config


def get_file(url, file_name=None):
    """Download a file from a URL and save it to a local cache directory.

        This function downloads a file from the specified URL and saves it to a local cache directory.
        If a local copy of the file with the same name already exists in the cache directory, it is not re-downloaded.

        Args:
            url (str): The URL of the file to be downloaded.
            file_name (str): The name to be used for the downloaded file.
                If None, the last part of the URL (after the last '/') is used as the file name.

        Returns:
            str: The path to the downloaded file in the local cache directory.
    """

    if file_name is None:
        file_name = url[url.rfind("/") + 1:]

    file_path = os.path.join(config.CACHE_DIR, file_name)

    if not os.path.exists(config.CACHE_DIR):
        os.mkdir(config.CACHE_DIR)

    if os.path.exists(file_path):
        return file_path

    print("Downloading: " + file_name)

    try:
        urllib.request.urlretrieve(url, file_path, _show_progress)

    except (Exception, KeyboardInterrupt):
        if os.path.exists(file_path):
            os.remove(file_path)
        raise

    print(" Done")

    return file_path


def _show_progress(block_num, block_size, total_size):
    bar_template = "\r[{}] {:.2f}%"

    downloaded = block_num * block_size
    percent = downloaded / total_size * 100
    indicator_num = int(downloaded / total_size * 30)

    percent = percent if percent < 100. else 100.
    indicator_num = indicator_num if indicator_num < 30 else 30

    indicator = "#" * indicator_num + "." * (30 - indicator_num)
    print(bar_template.format(indicator, percent), end="")



import os
import subprocess

from PIL import Image


def plot_dot_graph(output, verbose=True, to_file="graph.png"):
    """Plot a computational graph as a DOT graph and save it to an image file.

        This function generates a DOT graph representing the computational graph backward-reachable
        from the given output container.
        The DOT graph is then saved as an image file in the specified format.

        Args:
            output (:class:`marquetry.Container`): The output container of the computational graph to visualize.
            verbose (bool, optional): If True, includes detailed information
                about container names, shapes, and data types in the DOT graph.
                Default is True.
            to_file (str): The file path where the DOT graph image should be saved.
                Default is "graph.png".

        Returns:
            None
    """

    dot_graph = get_dot_graph(output.node, verbose)

    tmp_dir = os.path.join(os.path.expanduser("~"), ".module_tmp")
    if not os.path.exists(tmp_dir):
        os.mkdir(tmp_dir)

    graph_path = os.path.join(tmp_dir, "tmp_graph.dot")

    with open(graph_path, "w") as f:
        f.write(dot_graph)

    extension = os.path.splitext(to_file)[1][1:]
    cmd = 'dot {} -T {} -o {}'.format(graph_path, extension, to_file)
    subprocess.run(cmd, shell=True)

    img = Image.open(to_file)
    img.show()


def get_dot_graph(output, verbose=True):
    """Generates a graphviz DOT text of a computational graph.

        Build a graph of functions and containers backward-reachable from the output.
        To visualize a graphviz DOT text, you need the dot binary from the graph viz
        package (www.graphviz.org).
    """
    txt = ""
    funcs = []
    seen_set = set()

    def add_func(f):
        if f not in seen_set:
            funcs.append(f)
            seen_set.add(f)

    add_func(output.creator)
    txt += _dot_var(output, verbose)

    while funcs:
        func = funcs.pop()
        txt += _dot_func(func)
        for x in func.inputs:
            txt += _dot_var(x, verbose)

            if x.creator is not None:
                add_func(x.creator)

    return 'digraph g {\n' + txt + "}"


def _dot_var(v, verbose=False):
    dot_var = '{} [label="{}", color=orange, style=filled]\n'

    name = "" if v.name is None else v.name
    if verbose and v.data is not None:
        if v.name is not None:
            name += ": "
        name += str(v.shape) + " " + str(v.dtype)

    return dot_var.format(id(v), name)


def _dot_func(f):
    dot_func = '{} [label="{}", color=lightblue, style=filled, shape=box]\n'
    ret = dot_func.format(id(f), f.__class__.__name__)

    dot_edge = '{} -> {}\n'
    for x in f.inputs:
        ret += dot_edge.format(id(x), id(f))
    for y in f.outputs:
        ret += dot_edge.format(id(f), id(y()))

    return ret



import numpy as np

from marquetry import cuda_backend
from marquetry import Container
from marquetry import Layer
from marquetry import Parameter


class Embedding(Layer):
    """Embedding layer for text and categorical data.

        The Embedding layer is commonly used for encoding text data or categorical data.
        It maps discrete values (e.g., words or categories) to dense vectors,
        allowing neural networks to process such data effectively.

        Args:
            vocab_size (int): The size of the vocabulary, i.e., the number of unique words or categories.
            embed_size (int): The size of the embedding vectors.

        Attributes:
            w (marquetry.Parameter): The embedding weight matrix.

        Examples:
            >>> x = np.arange(0, 10)
            >>> x
            array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
            >>> vocab_size = 10000
            >>> embed_size = 100
            >>> embedding = Embedding(vocab_size, embed_size)
            >>> y = embedding(x)
            >>> y.shape
            (10, 100)

    """
    def __init__(self, vocab_size, embed_size):
        super().__init__()
        self.w = Parameter(np.random.randn(vocab_size, embed_size))

    def __call__(self, x):
        if cuda_backend.get_array_module(x) is not np:
            self.to_gpu()

        y = self.w[x]

        return y

    def set_embedding_vector(self, vector):
        """Set custom embedding vectors for the layer.

            After set the custom vector, this layer doesn't learn the parameter anymore
            because the custom vector should be pre-trained by word2vec or so.

            Args:
                vector (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                    The custom embedding vectors to set.

            Notes:
                vector:
                    If custom vector is used,
                    the custom vector shape should match the input `vector_size` and `embed_size`.
        """

        if self.w.shape != vector.shape:
            raise ValueError("embed_vector shape expected {}, but got {}"
                             .format(self.w.shape, vector.shape))

        if isinstance(vector, Container):
            vector = vector.data

        self.w.data = vector
        self._params = set()



from marquetry import cuda_backend
from marquetry import Function


class Clip(Function):
    """Clip the values of an array within a specified range.

        This class clips the values of the input array `x`
        so that they fall within the specified range [`x_min`, `x_max`].
        Values less than `x_min` are set to `x_min`,
        and values greater than `x_max` are set to `x_max`.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, x_min, x_max):
        self.x_min = x_min
        self.x_max = x_max

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)

        y = xp.clip(x, self.x_min, self.x_max)

        return y

    def backward(self, x, grad_y):
        x = x[0]

        mask = (x >= self.x_min) * (x <= self.x_max)
        grad_x = grad_y[0] * mask

        return grad_x


def clip(x, x_min, x_max):
    """Clip the values of an array within a specified range.

        This function clips the values of the input array `x`
        so that they fall within the specified range [`x_min`, `x_max`]. \n
        Values less than `x_min` are set to `x_min`,
        and values greater than `x_max` are set to `x_max`.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input array or container.
            x_min (float or int): The minimum value for clipping.
            x_max (float or int): The maximum value for clipping.

        Returns:
            :class:`marquetry.Container`: The result of clipping, with values in the range [`x_min`, `x_max`].

        Examples:
            >>> x = np.array([[-0.2, 10.9, 3.2], [-1.2,  4.2, 0]])
            >>> x
            array([[-0.2, 10.9,  3.2],
                   [-1.2,  4.2,  0. ]])
            >>> clip(x, 0.1, 5.)
            container([[0.1 5.  3.2]
                       [0.1 4.2 0.1]])
            >>> clip(x, -1., 9.)
            container([[-0.2  9.   3.2]
                       [-1.   4.2  0. ]])
    """

    return Clip(x_min, x_max)(x)



import numpy as np

from marquetry import cuda_backend
from marquetry import Container


def array_equal(a, b):
    """Check if the arrays are the same or not."""
    a = a.data if isinstance(a, Container) else a
    b = b.data if isinstance(b, Container) else b

    a, b = cuda_backend.as_numpy(a), cuda_backend.as_numpy(b)

    return np.array_equal(a, b)


import pandas as pd

from marquetry import Preprocess


class ColumnStandardize(Preprocess):
    """Column Standardization for Specified Columns in a DataFrame.

        ColumnStandardize is a preprocessing step used to standardize specified columns in a Pandas DataFrame.
        It scales the values in the selected columns to have a minimum of 0 and a maximum of 1.

        Args:
            std_target_column (list): A list of column names to be standardized.
            name (str): The name of the ColumnStandardize instance.
            is_train (bool): A boolean flag indicating whether
                this ColumnStandardize instance is for training or inference.
            skip_nan (bool): If True, columns containing NaN (null) values will be skipped during standardization.
                If False, NaN values will raise an error.
                Defaults to True.

        Examples:
            >>> import pandas as pd
            >>> from marquetry.preprocesses import ColumnStandardize
            >>> data = pd.DataFrame({'FeatureA': [1.0, 2.0, 3.0, 4.0], 'FeatureB': [10.0, 20.0, 30.0, 40.0]})
            >>> standardizer = ColumnStandardize(std_target_column=['FeatureA', 'FeatureB'], name='standardizer', is_train=True)
            >>> standardized_data = standardizer(data)
            >>> print(standardized_data)
               FeatureA  FeatureB
            0  1.000000  1.000000
            1  0.666667  0.666667
            2  0.333333  0.333333
            3 -0.000000 -0.000000

        Note:
            ColumnStandardize scales the values in the specified columns to have a minimum of 0 and a maximum of 1.
            It provides an option to skip columns containing NaN (null) values during standardization.
    """

    _label = "pre_cs"

    def __init__(self, std_target_column: list, name: str, is_train: bool, skip_nan: bool = True):
        super().__init__(name, is_train)
        self._std_target_column = std_target_column
        self._skip_nan = skip_nan

    def process(self, data: pd.DataFrame):
        """Process the input DataFrame by standardizing specified columns.

            Args:
                data (:class:`pandas.DataFrame`): The input DataFrame with columns to be standardized.

            Returns:
                pd.DataFrame: The DataFrame with specified columns standardized.

            Caution:
                Generally, process will be called by ``Marquetry`` core.
                Please call like the ``EXAMPLES``.
        """

        if len(self._std_target_column) == 0:
            return data

        if self._statistic_data is None:
            self._calc_statistic(data)

        self._validate_values()

        standardized_data = data.copy()

        for column in self._std_target_column:
            tmp_dict = self._statistic_data[column]

            standardized_data.loc[:, column] = (
                    (data.loc[:, column] - tmp_dict["min_value"]) / (tmp_dict["max_value"] - tmp_dict["min_value"]))

        return standardized_data

    def _validate_values(self):
        exist_statistic_columns = set([key for key, value in self._statistic_data.items() if value != {}])
        input_target_columns = set(self._std_target_column)

        if exist_statistic_columns != input_target_columns:
            raise ValueError("saved statistic data's target columns is {} but you input {}. "
                             .format(exist_statistic_columns, input_target_columns) + self._msg)

        return

    def _calc_statistic(self, data: pd.DataFrame):
        standardize_dict = {}

        max_data = data.loc[:, tuple(self._std_target_column)].max()
        min_data = data.loc[:, tuple(self._std_target_column)].min()

        for column in list(data.columns):
            if column not in self._std_target_column:
                standardize_dict[column] = {}
                continue

            if not self._skip_nan and data[column].isna().sum() != 0:
                raise ValueError("input data has null value, but it can't be skipped by user configure "
                                 "so the normalize can't be done expected.")

            tmp_max = min_data[column]
            tmp_min = max_data[column]

            tmp_dict = {
                "max_value": tmp_max,
                "min_value": tmp_min
            }

            standardize_dict[column] = tmp_dict

        self._save_statistic(standardize_dict)
        self._statistic_data = standardize_dict



from typing import Literal

import numpy as np
import pandas as pd

from marquetry import Preprocess


class LabelEncode(Preprocess):
    """Label Encoding for Categorical Data.

        LabelEncode is a preprocessing step used to encode categorical columns in a Pandas DataFrame into
        numerical labels.
        This is often done to prepare categorical data for machine learning models that require numerical input.

        Args:
            category_column (list): A list of column names to be label encoded.
            name (str): The name of the LabelEncode instance.
            is_train (bool): A boolean flag indicating whether this LabelEncode instance is for training or inference.
            specify_label (dict or None): A dictionary specifying custom labels for categories.
                The keys are column names, and the values are dictionaries mapping category values to label values.
                Defaults to None.
            treat_unknown (Literal["no_label", "encode_specify_value", "raise_error"]):
                How to handle unknown categories that are not found in the `specify_label` dictionary.

                Options are:

                - "no_label": Leave unknown categories unlabeled (default).
                - "encode_specify_value": Encode unknown categories with a specified label.
                - "raise_error": Raise an error if unknown categories are encountered.

            unknown_value (int): The label value to use for unknown categories
                when `treat_unknown` is set to "encode_specify_value".
                Defaults to -1.
            include_null (bool): Whether to include null or NaN values when calculating labels.
                Defaults to False.

        Examples:
            >>> import pandas as pd
            >>> from marquetry.preprocesses import LabelEncode
            >>> data = pd.DataFrame({'Category': ['A', 'B', 'A', 'C']})
            >>> label_encoder = LabelEncode(category_column=['Category'], name='label_encoder', is_train=True)
            >>> encoded_data = label_encoder(data)
            >>> print(encoded_data)
               Category
            0         0
            1         1
            2         0
            3         2

        Note:
            LabelEncode converts categorical columns into numerical labels.
            And, You can specify custom labels for categories using the `specify_label` parameter and,
            Unknown categories can be handled in different ways using the `treat_unknown` parameter.
    """

    _label = "pre_le"

    def __init__(self, category_column: list, name: str, is_train: bool, specify_label: dict = None,
                 treat_unknown: Literal["no_label", "encode_specify_value", "raise_error"] = "encode_specify_value",
                 unknown_value: int = -1, include_null=False):
        super().__init__(name, is_train)
        self._category_column = category_column
        self._include_null = include_null
        self._treat_unknown = treat_unknown
        self._unknown_value = unknown_value
        self._specify_relation_dict = specify_label if specify_label is not None else {}
        if not all(column in self._category_column for column in self._specify_relation_dict.keys()):
            raise ValueError("specify label include not exist label, this dataset has {}."
                             .format(self._category_column))

    def process(self, data: pd.DataFrame):
        """Process the input DataFrame by label encoding specified columns.

            Args:
                data (:class:`pandas.DataFrame`): The input DataFrame to be label encoded.

            Returns:
                pd.DataFrame: The label encoded DataFrame.

            Caution:
                Generally, process will be called by ``Marquetry`` core.
                Please call like the ``EXAMPLES``.

        """

        if len(self._category_column) == 0:
            return data

        type_change_dict = {key: str for key in self._category_column}
        data = data.astype(type_change_dict).replace("nan", np.nan)

        if self._statistic_data is None:
            self._calc_statistic(data)

        unknown_handler = self._validate_values(data)

        if unknown_handler and self._treat_unknown == "encode_specify_value":
            for column in self._category_column:
                unique_set = set(data[column])
                statistic_unique = set(self._statistic_data[column].keys())

                unknown_set = unique_set - statistic_unique
                if len(unknown_set) == 0:
                    continue

                unknown_dict = {unknown_key: self._unknown_value for unknown_key in list(unknown_set)}

                self._statistic_data[column].update(unknown_dict)

        labeled_data = data.replace(self._statistic_data)
        labeled_data = labeled_data[self._category_column]

        return labeled_data

    def _validate_values(self, data):
        exist_statistic_columns = set([key for key, value in self._statistic_data.items() if value != {}])
        input_target_columns = set(self._category_column)

        diff_statistic_target = False

        if exist_statistic_columns != input_target_columns:
            raise ValueError("saved statistic data's target columns is {} but you input {}. "
                             .format(exist_statistic_columns, input_target_columns) + self._msg)

        for column in self._category_column:
            unique_set = set(data[column])
            if not self._include_null:
                unique_set = set(unique_value for unique_value in unique_set if not pd.isna(unique_value))

            statistic_set = set(pd.Series(self._statistic_data[column]).keys())

            diff_statistic_target = True if unique_set != statistic_set else False

            if unique_set != statistic_set and self._treat_unknown == "raise_error":
                raise ValueError("statistic data doesn't have {} category in '{}' but the input has it. "
                                 .format(",".join(sorted(list(unique_set - statistic_set))), column) + self._msg)

        return diff_statistic_target

    def _calc_statistic(self, data: pd.DataFrame):
        replace_dict = {}

        for column in list(data.columns):
            if column not in self._category_column:
                replace_dict[column] = {}
                continue

            tmp_series = data[column]
            unique_set = set(tmp_series)

            if column in self._specify_relation_dict:
                tmp_dict = self._specify_relation_dict[column]

            else:
                if not self._include_null:
                    unique_set = [unique_value for unique_value in unique_set if not pd.isna(unique_value)]

                class_nums = list(range(len(unique_set)))

                tmp_dict = dict(zip(sorted(list(unique_set)), class_nums))

            if not len(set(unique_set) - set(tmp_dict.keys())) == 0:
                raise ValueError("{} has {} unique values but your specifying value is {}. "
                                 "If you want to specify label to this column, please cover all values."
                                 .format(column, len(unique_set), tmp_dict))

            replace_dict[column] = tmp_dict

        self._save_statistic(replace_dict)
        self._statistic_data = replace_dict



import numpy as np

from marquetry import utils
from marquetry import MachineLearning


class ClassificationTree(MachineLearning):
    """A decision tree classifier for classification tasks.

        The ClassificationTree class implements a decision tree for classifying data points based on their features.
        It is capable of training on labeled data and making predictions on new data points.

        Args:
            max_depth (int or None): The maximum depth of the decision tree.
                If None, the tree can grow indefinitely.
            min_split_samples (int or None): The minimum number of samples required to split a node.
                If None, defaults to 1.
            criterion (str): The criterion used for splitting nodes.
                Should be one of {"gini", "entropy"}.
            seed (int or None): The random seed for reproducibility.

        Attributes:
            tree (dict or None): The internal representation of the decision tree.

        Methods:
            fit(self, x, t): Train the decision tree on the input data and target labels.
            predict(self, x): Make predictions on new data points.
            score(self, x, t, evaluator): Evaluate the performance of the decision tree using an evaluator function.

        Examples:
            >>> model = ClassificationTree(max_depth=5, criterion="entropy", seed=42)
            >>> model.fit(training_data, training_labels)
            >>> predictions = model.predict(new_data)
            >>> model.tree
            {
                'feature': 2,
                'threshold': 2.45,
                'content': 'branch',
                'left_branch': ...,
                'right_branch':...,
            }

    """

    _expect_criterion = ("gini", "entropy")

    def __init__(self, max_depth=None, min_split_samples=None, criterion="gini", seed=None):
        if criterion not in self._expect_criterion:
            raise ValueError("criterion expects {}, but got {}".
                             format(",".join(self._expect_criterion), criterion))

        self.criterion = criterion
        self.seed = seed
        self.max_depth = max_depth if max_depth is not None else float("inf")
        self.min_split_samples = min_split_samples if min_split_samples is not None else 1

        self.tree = None
        self.unique_list = None

    def _fit_method(self, x, t):

        self.unique_list = np.unique(t).tolist()
        self.tree = self._recurrent_create_tree(x, t, 0)

    def _predict_method(self, x):

        predict_result = []

        for sample in x:
            predict_result.append(self._recurrent_prediction(sample, self.tree))

        return np.array(predict_result)

    def _recurrent_create_tree(self, x, t, depth):
        is_leaf = True if len(x) < self.min_split_samples or depth == self.max_depth else False
        is_leaf, (label, impurity), feature, threshold = (
            utils.split_branch(x, t, self.unique_list, criterion=self.criterion, seed=self.seed, is_leaf=is_leaf))

        if is_leaf:
            tmp_dict = {
                "content": "leaf",
                "label": label,
                "train_impurity": impurity
            }

        else:
            x_left = x[x[:, feature] <= threshold]
            t_left = t[x[:, feature] <= threshold]
            x_right = x[x[:, feature] > threshold]
            t_right = t[x[:, feature] > threshold]

            tmp_dict = {
                "feature": feature,
                "threshold": threshold,
                "content": "branch",
                "left_branch": self._recurrent_create_tree(x_left, t_left, depth + 1),
                "right_branch": self._recurrent_create_tree(x_right, t_right, depth + 1)
            }

        return tmp_dict

    def _recurrent_prediction(self, x, tree: dict):
        variable = tree["content"]

        if variable == "branch":
            threshold = tree["threshold"]
            feature = tree["feature"]
            if x[feature] <= threshold:
                return self._recurrent_prediction(x, tree["left_branch"])
            else:
                return self._recurrent_prediction(x, tree["right_branch"])
        elif variable == "leaf":
            return tree["label"]
        else:
            raise Exception("Something internal implement wrong please notify this to the developer.")



def sum_to(x, shape):
    """Sum elements along axes to output an array of a given shape."""
    ndim = len(shape)
    lead = x.ndim - ndim
    lead_axis = tuple(range(lead))

    axis = tuple([i + lead for i, sx in enumerate(shape) if sx == 1])
    y = x.sum(lead_axis + axis, keepdims=True)
    if lead > 0:
        y = y.squeeze(lead_axis)
    return y



import json
import os
import warnings

import pandas as pd

from marquetry import configuration


# ===========================================================================
# preprocess base class
# ===========================================================================
class Preprocess(object):
    """Base class for preprocessing CSV or Datamart(Not implement now) data.

        All preprocess implementation defined in :mod:`marquetry.preprocesses` inherit
        this class.

        The main feature of this class is to provide a uniform process for all preprocessing steps.
        When a preprocess function receives input data, it first checks the data type and structure.
        After that, it resets the data index and performs the preprocessing.

        Attributes:
            data_dir (str): The directory where statistic data is stored.

        Args:
            name (str): A unique name for the preprocess instance.
                It is used for saving and loading statistic data.
            is_train(bool): Specifying the data is ``train`` or not.

    """

    _label = None
    _msg = """if you use new data for the training, please use new `name` parameter or delete the old statistic data"""

    def __init__(self, name, is_train):
        self._name = name
        self._is_train = is_train

        data_dir = os.path.join(configuration.config.CACHE_DIR, name)
        if not os.path.exists(data_dir):
            os.makedirs(data_dir)

        self.data_dir = data_dir
        self._statistic_data, self._statistic_size = self._load_statistic()

    def __call__(self, data: pd.DataFrame):
        if not isinstance(data, pd.DataFrame):
            raise TypeError("Preprocess requires the input is pandas DataFrame.")

        if self._statistic_data is not None:
            self._validate_structure(data)

        if self._statistic_data is None and not self._is_train:
            raise ValueError("In test data process, train statistic must be needed but there is no statistic data.")

        if self._statistic_size is None:
            self._statistic_size = len(data.index)

        data.reset_index(drop=True, inplace=True)
        output = self.process(data)

        if isinstance(output, tuple):
            output = output[0]

        return output

    def process(self, data):
        """Define perform custom preprocessing on the input data. (to be implemented by subclasses)

            Args:
                data (pd.DataFrame): Input data in the form of a pandas DataFrame.

            Returns:
                pd.DataFrame: Preprocessed data.

        """

        raise NotImplementedError()

    def _validate_structure(self, data: pd.DataFrame):
        """Validate the structure of the input data compared to saved statistic data.

            Args:
                data (pd.DataFrame): Input data in the form of a pandas DataFrame.

        """
        if self._statistic_size != len(data) and self._is_train:
            self._statistic_data = None
            self.remove_old_statistic()
            self._statistic_size = len(data.index)

            return

        if not isinstance(self._statistic_data, dict):
            msg = "statistic data is wrong, expected dict but got {}. ".format(type(self._statistic_data))
            if not self._is_train:
                raise TypeError(msg)

            warnings.warn(msg + "Delete and Recalculate the statistic file.", SyntaxWarning)
            self._statistic_data = None
            self.remove_old_statistic()

            return

        data_columns = set(data.columns)
        statistic_columns = set(self._statistic_data.keys())

        if data_columns != statistic_columns:
            raise ValueError("Statistic data unmatch with input data. Please check correctness of the statistic name.")

        return

    def _save_statistic(self, statistic_data: dict):
        """Save statistic data to a file.

            Args:
                statistic_data (dict): Statistic data to be saved.

            Note:
                This method requires self._label is not None.
                Therefore, if this method call in the base class, always it failed by NotImplementedError.
        """

        if self._label is None:
            raise NotImplementedError()

        file_name = self._name + "." + self._label + ".json"
        file_path = os.path.join(self.data_dir, file_name)

        save_data_dict = {"statistic_data": statistic_data, "statistic_size": self._statistic_size}

        with open(file_path, "w") as f:
            json.dump(save_data_dict, f)

        return

    def _load_statistic(self):
        """Load saved statistic data from a file.

            Returns:
                dict or None: Loaded statistic data, or None if no data is found.

            Note:
                This method requires self._label is not None.
                Therefore, if this method call in the base class, always it failed by NotImplementedError.

        """

        if self._label is None:
            raise NotImplementedError()

        file_name = self._name + "." + self._label + ".json"
        file_path = os.path.join(self.data_dir, file_name)

        if not os.path.exists(file_path):
            return None, None

        with open(file_path, "r") as f:
            load_data = json.load(f)

        statistic_data = load_data["statistic_data"]
        data_size = load_data["statistic_size"]

        return statistic_data, data_size

    def remove_old_statistic(self):
        """Remove previously saved statistic data."""

        file_name = self._name + "." + self._label + ".json"
        file_path = os.path.join(self.data_dir, file_name)

        if os.path.exists(file_path):
            os.remove(file_path)

        self._statistic_data = None

        return



from marquetry.optimizers import ada_grad
from marquetry.optimizers import adam
from marquetry.optimizers import momentum_sgd
from marquetry.optimizers import rmsprop
from marquetry.optimizers import sgd


from marquetry.optimizers.ada_grad import AdaGrad
from marquetry.optimizers.adam import Adam
from marquetry.optimizers.momentum_sgd import MomentumSGD
from marquetry.optimizers.rmsprop import RMSProp
from marquetry.optimizers.sgd import SGD



import numpy as np

from marquetry import cuda_backend
from marquetry import Function
from marquetry import functions


class Split(Function):
    """Split an input array or container into multiple parts along the specified axis and indices.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, indices, axis):
        self.axis = axis
        if np.isscalar(indices):
            indices = (indices,)
        self.indices = indices

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.split(x, self.indices, axis=self.axis)

        self.retain_inputs(())
        return tuple(y)

    def backward(self, x, grad_ys):
        grad_x = functions.concat(grad_ys, axis=self.axis)

        return grad_x


def split(x, indices, axis):
    """Split an input array or container into multiple parts along the specified axis and indices.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input array to be split.
            indices (int or tuple of ints):
                The indices at which to split the input array or container along the specified axis.
            axis (int): The axis along which the input array or container should be split.

        Returns:
            list of :class:`marquetry.Container` a tuple containing the result of
            splitting the input array into multiple parts along the specified axis.

        Examples:
            >>> x = np.arange(1, 9).reshape(2, 4)
            array([[1, 2, 3, 4],
                   [5, 6, 7, 8]])
            >>> split(x, indices=(2, 3), axis=1)
            [container([[1 2]
                        [5 6]]),
             container([[3]
                        [7]]),
             container([[4]
                        [8]])]
    """

    return Split(indices, axis)(x)



import numpy as np


class AsType(object):
    """A utility class for converting NumPy arrays to a specific data type.

        Args:
            dtype (str or numpy.dtype):
                The target data type to which arrays should be converted.

        Examples:
            >>> MyDataset(transform=AsType(np.float32))
            the transform function is applied to the dataset when extract the data.
    """

    def __init__(self, dtype=np.float32):
        self.dtype = dtype

    def __call__(self, array):
        return array.astype(self.dtype)


class ToFloat(AsType):
    """
        A subclass of AsType for converting NumPy arrays to float data type.
        This class inherits from AsType and sets the target data type to float32.

        Examples:
            >>> MyDataset(transform=ToFloat())
    """

    def __init__(self):
        super().__init__(dtype=np.float32)


class ToInt(AsType):
    """
        A subclass of AsType for converting NumPy arrays to integer data type.
        This class inherits from AsType and sets the target data type to int32.

        Examples:
            >>> MyDataset(transform=ToInt())
    """

    def __init__(self):
        super().__init__(dtype=np.int32)



from marquetry import cuda_backend
from marquetry import Function
from marquetry import functions


class Squeeze(Function):
    """Remove single-dimensional entries from the shape of an input array.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def __init__(self, axis):
        if isinstance(axis, int):
            axis = (axis,)
        elif isinstance(axis, (tuple, list)) and all(isinstance(x, int) for x in axis):
            axis = tuple(axis)
        elif axis is None:
            axis = axis
        else:
            raise ValueError("axis expected int or sequence of ints, but got invalid type data.")

        self.axis = axis

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)

        if self.axis is not None:
            for axis in self.axis:
                if x.shape[axis] != 1:
                    raise ValueError("You can't squeeze non-one size axis element, {}-axis is {}."
                                     .format(axis, x.shape[axis]))

        y = xp.squeeze(x, axis=self.axis)

        return y

    def backward(self, x, grad_y):
        axis = self.axis
        if axis is None:
            tmp_axis = []
            for i, axis_num in enumerate(x[0].shape):
                if axis_num == 1:
                    tmp_axis.append(i)

            axis = tuple(tmp_axis)

        grad_x = functions.unsqueeze(grad_y[0], axis)

        return grad_x


def squeeze(x, axis):
    """Remove single-dimensional entries from the shape of an input array.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input array  from which single-dimensional entries should be removed.
            axis (int or tuple of ints or list of ints): The axis along which single-dimensional entries should be removed.

        Returns:
            :class:`marquetry.Container`: The result of removing single-dimensional
            entries from the input array or container along the specified axis (or all axes if axis is None).

        Raises:
            ValueError: If the specified axis does not have a size of 1,
                indicating that there are no single-dimensional entries to remove.

        Examples:
            >>> x = np.arange(1, 9).reshape(1, 8)
            array([[1, 2, 3, 4, 5, 6, 7, 8]])
            >>> squeeze(x, axis=0)
            container([1 2 3 4 5 6 7 8])

    """

    return Squeeze(axis)(x)



from marquetry import Function
from marquetry import functions
from marquetry import utils


class Col2im(Function):
    """Export image array from array.

        This class is commonly used by the backward computation of the convolution.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, image_shape, kernel_size, stride, pad, to_matrix):
        super().__init__()

        self.image_shape = image_shape
        self.kernel_size = kernel_size
        self.stride = stride
        self.pad = pad
        self.to_matrix = to_matrix

    def forward(self, x):
        y = utils.col2im_array(x, self.image_shape, self.kernel_size, self.stride, self.pad, self.to_matrix)

        self.retain_inputs(())
        return y

    def backward(self, grad_y):
        grad_x = functions.im2col(grad_y[0], self.kernel_size, self.stride, self.pad, self.to_matrix)

        return grad_x


def col2im(col, image_shape, kernel_size, stride=1, pad=0, to_matrix=True):
    """Export image array from array.

        This function is used by the backward computation of the convolution.

        Args:
            col (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                Input container that is :class:`marquetry.Container` array.
            image_shape (tuple): image shape the output image.
            kernel_size (int, tuple(y, x)): The convolution process kernel size.
            stride (int, tuple(y, x)): The convolution process stride size
            pad (int, tuple): The convolution process padding size
            to_matrix (bool): If this ``True``, the input array should be 2-dim array.
                Otherwise, the input array should be 6-dims array
                (batch_size, channels, kernel_height, kernel_width, output_height, output_width).
    """

    return Col2im(image_shape, kernel_size, stride, pad, to_matrix)(col)



from marquetry import cuda_backend
from marquetry import Function


class MeanAbsoluteError(Function):
    """Calculate the Mean Absolute Error (MAE) between predicted values and true values.

        This class defines that calculates the Mean Absolute Error (MAE) between predicted values and true Values.
        MAE measures the average absolute difference between the predicted and true values.
        It can be calculated either as a uniform average over all samples or as raw values for each column.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def __init__(self, multi_output):
        if multi_output in ["uniform_average", "raw_values"]:
            self.multi_output = multi_output
        else:
            raise ValueError("invalid multi_output argument")

    def forward(self, y, t):
        assert y.size == t.size

        xp = cuda_backend.get_array_module(y)
        if t.ndim == 1:
            t = t.reshape((-1, 1))
        y = y.reshape(t.shape)

        mae_value = xp.mean(xp.absolute(y - t), axis=0)

        self.retain_inputs(())

        if self.multi_output == "uniform_average":
            return xp.asarray(mae_value.mean(), dtype=y.dtype)
        else:
            return xp.asarray(mae_value, dtype=y.dtype)


def mean_absolute_error(y, t, multi_output="uniform_average"):
    """Calculate the Mean Absolute Error (MAE) between predicted values and true values.

        This function defines that calculates the Mean Absolute Error (MAE) between predicted values and true Values.
        MAE measures the average absolute difference between the predicted and true values.
        It can be calculated either as a uniform average over all samples or as raw values for each column.

        Args:
            y (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted values.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true values.
            multi_output (str): Specifies how to calculate the MAE for multi-output.

        Note:
            multi_output:
                "uniform_average": Compute the uniform average MAE over all samples. This is the default option.

                "raw_values": Return the raw MAE values for each sample.

        Returns:
            :class:`marquetry.Container`: The MAE values based on the predicted values
                and true values.
    """
    return MeanAbsoluteError(multi_output)(y, t)


mae = mean_absolute_error



import marquetry
from marquetry import functions
from marquetry import Layer
from marquetry.layers import Linear


class LSTM(Layer):
    """Long Short-Term Memory (LSTM) layer for sequence modeling.

        The LSTM layer is a type of recurrent neural network (RNN) layer used for modeling sequential data.
        It is designed to capture long-term dependencies in sequential data and is a variation of the standard RNN.

        Args:
            hidden_size (int): The size of the hidden state.
            in_size (int or None): The size of the input data.

        Caution:
            in_size:
                This is automatically determined from the input data shape
                and does not need to be specified except a special use case.

        Attributes:
            h (:class:`marquetry.Container`): The current hidden state.
            c (:class:`marquetry.Container`): The current cell state.

        Examples:
            >>> dataset = marquetry.datasets.SinCurve()
            >>> dataloader = marquetry.dataloaders.SeqDataLoader(dataset, batch_size=32)
            >>> model = (LSTM(128), marquetry.layers.Linear(1))
            >>> loss = 0
            >>> for x, t in dataloader:
            >>>     for layer in model:
            >>>         x = layer(x)
            >>>     loss += functions.mean_squared_error(x, t)
            >>> loss
            container(28.759597353348106)

    """

    def __init__(self, hidden_size, in_size=None):
        super().__init__()

        self.hidden_size = hidden_size

        self.x2hs = Linear(3 * hidden_size, in_size=in_size)
        self.x2i = Linear(hidden_size, in_size=in_size)
        self.h2hs = Linear(3 * hidden_size, in_size=hidden_size, nobias=True)
        self.h2i = Linear(hidden_size, in_size=hidden_size, nobias=True)

        self.h = None
        self.c = None

    def reset_state(self):
        """Reset the hidden state and cell state."""
        self.h = None
        self.c = None

    def set_state(self, h, c=None):
        """Set the hidden state and cell state to a custom value.

            Args:
                h (marquetry.Container): The custom hidden state.
                c (marquetry.Container or None, optional): The custom cell state.

            Caution:
                Almost general use case, the cell state should NOT set custom value
                because cell state in LSTM is used only internal information connection,
                and it should be managed automatically.
                If you don't have any special reason, you should set only hidden state.

        """
        if not isinstance(h, marquetry.Container):
            raise TypeError("hidden state type should be marquetry.Container, but got {}."
                            .format(type(h).__name__))

        if c is not None and not isinstance(c, marquetry.Container):
            raise TypeError("cell state type should be (marquetry.Container or None), but got {}."
                            .format(type(c).__name__))

        self.h = h
        if c is not None:
            self.c = c

    def forward(self, x):
        if self.h is None:
            hs = functions.sigmoid(self.x2hs(x))
            input_data = functions.tanh(self.x2i(x))
        else:
            hs = functions.sigmoid(self.x2hs(x) + self.h2hs(self.h))
            input_data = functions.tanh(self.x2i(x) + self.h2i(self.h))

        forget_gate = hs[:, :self.hidden_size]
        input_gate = hs[:, self.hidden_size:2 * self.hidden_size]
        output_gate = hs[:, 2 * self.hidden_size:]

        if self.c is None:
            c_new = input_gate * input_data
        else:
            c_new = (forget_gate * self.c) + (input_gate * input_data)

        h_new = output_gate * functions.tanh(c_new)

        self.h, self.c = h_new, c_new

        return h_new



from marquetry import cuda_backend
from marquetry import Function
from marquetry import functions
from marquetry import utils


class SoftmaxCrossEntropy(Function):
    """Calculate the Softmax Cross-Entropy loss between predicted values and true labels.

        This function calculates the Softmax Cross-Entropy loss,
        which measures the difference between the predicted values (scores) and
        the true labels for multi-class classification tasks.
        It takes the softmax of the input values to compute class probabilities and then
        compares them to the true labels.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def forward(self, x, t):
        xp = cuda_backend.get_array_module(x)
        batch_size = x.shape[0]
        log_z = utils.log_sum_exp(x, axis=1)
        log_p = x - log_z
        log_p = log_p[xp.arange(batch_size), t.ravel()]
        y = -log_p.sum() / xp.float32(batch_size)

        return y

    def backward(self, inputs, grad_y):
        x, t = inputs
        grad_y = grad_y[0]

        batch_size, data_dim = x.shape

        grad_y *= 1 / batch_size
        y = functions.softmax(x)
        xp = cuda_backend.get_array_module(t)
        if y.size != t.size:
            # convert class num to one-hot
            t_onehot = xp.eye(data_dim, dtype=t.dtype)[t]
        else:
            t_onehot = t

        y = (y - t_onehot) * grad_y
        return y


def softmax_cross_entropy(x, t):
    """Calculate the Softmax Cross-Entropy loss between predicted values and true labels.


        This function calculates the Softmax Cross-Entropy loss,
        which measures the difference between the predicted values(scores) and
        the true labels for multi-class classification tasks.
        It takes the softmax of the input values to compute class probabilities and then
        compares them to the true labels.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted values or scores.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true labels.

        Returns:
            :class:`marquetry.Container`: The Softmax Cross-Entropy loss.
    """

    return SoftmaxCrossEntropy()(x, t)



from marquetry import Function
from marquetry import functions


class UnSqueeze(Function):
    """Add a single-dimensional entry (axis with size 1) belong to the specify axis to the shape of an input array.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, axis):
        if isinstance(axis, int):
            axis = (axis,)
        elif isinstance(axis, (tuple, list)) and all(isinstance(x, int) for x in axis):
            axis = tuple(axis)
        else:
            raise ValueError("axis expected int or sequence of ints, but got invalid type data.")

        self.axis = axis

    def forward(self, x):
        x_shape = x.shape

        new_shape = list(x_shape)

        for index in sorted(self.axis):
            new_shape.insert(index, 1)
        new_shape = tuple(new_shape)

        y = x.reshape(new_shape)

        self.retain_inputs(())
        return y

    def backward(self, x, grad_y):
        grad_x = functions.squeeze(grad_y[0], self.axis)

        return grad_x


def unsqueeze(x, axis):
    """Add a single-dimensional entry (axis with size 1) belong to the specify axis to the shape of an input array.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input array to which a single-dimensional entry should be added along the specified axis.
            axis (int or tuple of ints or list of ints): The axis along which to add the single-dimensional entry.

        Returns:
            :class:`marquetry.Container`:
                The result of adding a single-dimensional entry to the input array along the specified axis.

        Examples:
            >>> x = np.arange(1, 9).reshape(1, 8)
            array([[1, 2, 3, 4, 5, 6, 7, 8]])
            >>> unsqueeze(x, axis=0)
            container([[[1 2 3 4 5 6 7 8]]])
            >>> unsqueeze(x, axis=0).shape
            (1, 1, 8)
        """

    return UnSqueeze(axis)(x)



from marquetry import utils
from marquetry.ml import ClassificationTree


class RegressionTree(ClassificationTree):
    """A decision tree regressor for predicting numerical targets.

        The RegressionTree class implements a decision tree for predicting numerical targets based on input features.
        It is capable of training on labeled data with numerical targets and making predictions on new data points.

        Args:
            max_depth (int or None): The maximum depth of the decision tree.
                If None, the tree can grow indefinitely.
            min_split_samples (int or None): The minimum number of samples required to split a node.
                If None, defaults to 1.
            criterion (str): The criterion used for splitting nodes.
                Should be one of {"rss", "mae"}.
            seed (int or None): The random seed for reproducibility.

        Attributes:
            tree (dict or None): The internal representation of the decision tree.

        Methods:
            fit(self, x, t): Train the decision tree on the input data and target labels.
            predict(self, x): Make predictions on new data points.
            score(self, x, t, evaluator): Evaluate the performance of the decision tree using an evaluator function.

        Examples:
            >>> model = RegressionTree(max_depth=5, criterion="rss", seed=42)
            >>> model.fit(training_data, training_labels)
            >>> predictions = model.predict(new_data)
            >>> model.tree
            {
                'feature': 8,
                'threshold': -0.00016962857797942404,
                'content': 'branch',
                'left_branch': ...,
                'right_branch': ...,
            }

    """

    _expect_criterion = ("rss", "mae")

    def __init__(self, max_depth=None, min_split_samples=None, criterion="rss", seed=None):
        super().__init__(max_depth, min_split_samples, criterion, seed)

    def _fit_method(self, x, t):
        self.tree = self._recurrent_create_tree(x, t, 0)

    def _recurrent_create_tree(self, x, t, depth):
        is_leaf = True if len(x) < self.min_split_samples or depth == self.max_depth else False
        is_leaf, (value, impurity), feature, threshold = (
            utils.split_branch(x, t, criterion=self.criterion,
                               seed=self.seed, target_type="regression", is_leaf=is_leaf))

        if is_leaf:
            tmp_dict = {
                "content": "leaf",
                "label": value,
                "train_impurity": impurity
            }

        else:
            x_left = x[x[:, feature] <= threshold]
            t_left = t[x[:, feature] <= threshold]
            x_right = x[x[:, feature] > threshold]
            t_right = t[x[:, feature] > threshold]

            tmp_dict = {
                "feature": feature,
                "threshold": threshold,
                "content": "branch",
                "left_branch": self._recurrent_create_tree(x_left, t_left, depth + 1),
                "right_branch": self._recurrent_create_tree(x_right, t_right, depth + 1)
            }

        return tmp_dict



import unittest

import numpy as np
from sklearn import metrics

from marquetry import functions
from marquetry.utils import array_close


class TestAccuracy(unittest.TestCase):

    def test_forward1(self):
        y = np.array([[0.8, 0.1, 0.05, 0.05], [0.1, 0.01, 0.59, 0.3], [0.0, 0.01, 0.32, 0.67]])
        t = np.array([0, 2, 2])

        expected = np.array(2 / 3)
        accuracy_score = functions.evaluation.accuracy(y, t)

        self.assertEqual(expected, accuracy_score.data)

    def test_forward2(self):
        y = np.array([
            [0.8, 0.1, 0.05, 0.05],
            [0.1, 0.01, 0.59, 0.3],
            [0.0, 0.01, 0.32, 0.67],
            [0.9, 0.03, 0.03, 0.04]])
        t = np.array([0, 3, 3, 1])

        ignore_class = 1

        expected = np.array(2 / 3)
        accuracy_score = functions.evaluation.accuracy(y, t, ignore_class)

        self.assertEqual(expected, accuracy_score.data)

    def test_forward3(self):
        y = np.array([0.89, 0.03, 0.56, 0.61, 0.65, 0.98])
        t = np.array([1, 0, 0, 1, 1, 1])

        expected = np.array(4 / 6)
        accuracy_score = functions.evaluation.binary_accuracy(y, t)

        self.assertEqual(expected, accuracy_score.data)

    def test_forward4(self):
        y = np.array([0.89, 0.03, 0.56, 0.61, 0.65, 0.98])
        t = np.array([1, 0, 0, 1, 1, 1])

        expected = np.array(1.)
        accuracy_score = functions.evaluation.binary_accuracy(y, t, threshold=0.6)

        self.assertEqual(expected, accuracy_score.data)


class TestFScore(unittest.TestCase):

    def test_forward1(self):
        y = np.array([0.89, 0.03, 0.56, 0.61, 0.65, 0.98])
        t = np.array([1, 0, 0, 1, 1, 1])

        y_pred = np.asarray(y >= .7, dtype="i")

        expected = metrics.f1_score(t, y_pred)
        f_score = functions.evaluation.f_score(y, t)

        self.assertEqual(expected, f_score.data)

    def test_forward2(self):
        y = np.array([0.89, 0.03, 0.56, 0.61, 0.65, 0.98])
        t = np.array([0, 0, 0, 0, 0, 0])

        y_pred = np.asarray(y >= .7, dtype="i")

        expected = metrics.f1_score(t, y_pred)
        f_score = functions.evaluation.f_score(y, t)

        self.assertEqual(expected, f_score.data)

    def test_forward3(self):
        y = np.array([
            [0.8, 0.1, 0.05, 0.05],
            [0.1, 0.01, 0.59, 0.3],
            [0.0, 0.01, 0.32, 0.67],
            [0.9, 0.03, 0.03, 0.04]])
        t = np.array([0, 3, 3, 1])

        target_class = 3
        y_pred = np.argmax(y, axis=1)

        pred_trans = np.asarray(y_pred == target_class, dtype="i")
        correct_trans = np.asarray(t == target_class, dtype="i")

        expected = metrics.f1_score(correct_trans, pred_trans)
        multi_f_score = functions.evaluation.multi_f_score(y, t, target_class=target_class)

        self.assertEqual(expected, multi_f_score.data)


class TestClassificationError(unittest.TestCase):

    def test_forward1(self):
        y = np.array([
            [0.8, 0.1, 0.05, 0.05],
            [0.1, 0.01, 0.59, 0.3],
            [0.0, 0.01, 0.32, 0.67],
            [0.9, 0.03, 0.03, 0.04]])
        t = np.array([0, 3, 3, 1])

        expected = np.array(2 / 4)
        cl_error = functions.evaluation.classification_error(y, t)

        self.assertEqual(expected, cl_error.data)

    def test_forward2(self):
        y = np.array([
            [0.8, 0.1, 0.05, 0.05],
            [0.1, 0.01, 0.59, 0.3],
            [0.0, 0.01, 0.32, 0.67],
            [0.9, 0.03, 0.03, 0.04]])
        t = np.array([0, 3, 3, 1])

        ignore_label = 1

        expected = np.array(1 / 3)
        cl_error = functions.evaluation.classification_error(y, t, ignore_label)

        self.assertEqual(expected, cl_error.data)

    def test_forward3(self):
        y = np.array([0.89, 0.03, 0.56, 0.61, 0.65, 0.98])
        t = np.array([1, 0, 0, 1, 1, 1])

        expected = np.array(1 / 3)
        cl_error = functions.evaluation.binary_classification_error(y, t)

        self.assertEqual(expected, cl_error.data)

    def test_forward4(self):
        y = np.array([0.89, 0.03, 0.56, 0.61, 0.65, 0.98])
        t = np.array([1, 0, 0, 1, 1, 1])

        expected = np.array(0 / 6)
        cl_error = functions.evaluation.binary_classification_error(y, t, threshold=0.6)

        self.assertEqual(expected, cl_error.data)


class TestPrecision(unittest.TestCase):

    def test_forward1(self):
        y = np.array([0.89, 0.03, 0.56, 0.61, 0.65, 0.98])
        t = np.array([1, 0, 0, 1, 1, 1])

        y_pred = np.asarray(y >= .7, dtype="i")

        expected = metrics.precision_score(t, y_pred)
        precision = functions.evaluation.precision(y, t)

        self.assertEqual(expected, precision.data)

    def test_forward2(self):
        y = np.array([0.89, 0.03, 0.56, 0.61, 0.65, 0.98])
        t = np.array([0, 0, 0, 0, 0, 0])

        y_pred = np.asarray(y >= .7, dtype="i")

        expected = metrics.precision_score(t, y_pred)
        precision = functions.evaluation.precision(y, t)

        self.assertEqual(expected, precision.data)

    def test_forward3(self):
        y = np.array([
            [0.8, 0.1, 0.05, 0.05],
            [0.1, 0.01, 0.59, 0.3],
            [0.0, 0.01, 0.32, 0.67],
            [0.9, 0.03, 0.03, 0.04]])
        t = np.array([0, 3, 3, 1])

        target_class = 3
        y_pred = np.argmax(y, axis=1)

        pred_trans = np.asarray(y_pred == target_class, dtype="i")
        correct_trans = np.asarray(t == target_class, dtype="i")

        expected = metrics.precision_score(correct_trans, pred_trans)
        multi_precision = functions.evaluation.multi_precision(y, t, target_class=target_class)

        self.assertEqual(expected, multi_precision.data)


class TestRecall(unittest.TestCase):

    def test_forward1(self):
        y = np.array([0.89, 0.03, 0.56, 0.61, 0.65, 0.98])
        t = np.array([1, 0, 0, 1, 1, 1])

        y_pred = np.asarray(y >= .7, dtype="i")

        expected = metrics.recall_score(t, y_pred)
        recall = functions.evaluation.recall(y, t)

        self.assertEqual(expected, recall.data)

    def test_forward2(self):
        y = np.array([0.89, 0.03, 0.56, 0.61, 0.65, 0.98])
        t = np.array([0, 0, 0, 0, 0, 0])

        expected = np.array(0.)
        recall = functions.evaluation.recall(y, t)

        self.assertEqual(expected, recall.data)

    def test_forward3(self):
        y = np.array([
            [0.8, 0.1, 0.05, 0.05],
            [0.1, 0.01, 0.59, 0.3],
            [0.0, 0.01, 0.32, 0.67],
            [0.9, 0.03, 0.03, 0.04]])
        t = np.array([0, 3, 3, 1])

        target_class = 3
        y_pred = np.argmax(y, axis=1)

        pred_trans = np.asarray(y_pred == target_class, dtype="i")
        correct_trans = np.asarray(t == target_class, dtype="i")

        expected = metrics.recall_score(correct_trans, pred_trans)
        multi_recall = functions.evaluation.multi_recall(y, t, target_class=target_class)

        self.assertEqual(expected, multi_recall.data)


class TestMAE(unittest.TestCase):

    def test_forward1(self):
        y = np.array([1001.2, 211.3, 4500.1, 200.0])
        t = np.array([1231., 199.2, 4328., 176.])

        sk_mae = metrics.mean_absolute_error(t, y)
        mae = functions.evaluation.mean_absolute_error(y, t)

        self.assertEqual(sk_mae, mae.data)

    def test_forward2(self):
        y = np.array([[1001.2, 211.3, 4500.1, 200.0], [1421.2, 231.3, 40.1, 140.0]])
        t = np.array([[1231., 199.2, 4328., 176.], [1312., 220.2, 43.4, 116.]])

        sk_mae = metrics.mean_absolute_error(t, y, multioutput="raw_values")
        mae = functions.evaluation.mean_absolute_error(y, t, multi_output="raw_values")

        self.assertTrue(array_close(sk_mae, mae.data))

    def test_forward3(self):
        y = np.array([[1001.2, 211.3, 4500.1, 200.0], [1421.2, 231.3, 40.1, 140.0]])
        t = np.array([[1231., 199.2, 4328., 176.], [1312., 220.2, 43.4, 116.]])

        sk_mae = metrics.mean_absolute_error(t, y)
        mae = functions.evaluation.mean_absolute_error(y, t)

        self.assertTrue(array_close(sk_mae, mae.data))


class TestMSE(unittest.TestCase):

    def test_forward1(self):
        y = np.array([1001.2, 211.3, 4500.1, 200.0])
        t = np.array([1231., 199.2, 4328., 176.])

        sk_mse = metrics.mean_squared_error(t, y)
        mse = functions.evaluation.mean_squared_error(y, t)

        self.assertEqual(sk_mse, mse.data)

    def test_forward2(self):
        y = np.array([[1001.2, 211.3, 4500.1, 200.0], [1421.2, 231.3, 40.1, 140.0]])
        t = np.array([[1231., 199.2, 4328., 176.], [1312., 220.2, 43.4, 116.]])

        sk_mse = metrics.mean_squared_error(t, y, multioutput="raw_values")
        mse = functions.evaluation.mean_squared_error(y, t, multi_output="raw_values")

        self.assertTrue(array_close(sk_mse, mse.data))

    def test_forward3(self):
        y = np.array([[1001.2, 211.3, 4500.1, 200.0], [1421.2, 231.3, 40.1, 140.0]])
        t = np.array([[1231., 199.2, 4328., 176.], [1312., 220.2, 43.4, 116.]])

        sk_mse = metrics.mean_squared_error(t, y)
        mse = functions.evaluation.mean_squared_error(y, t)

        self.assertTrue(array_close(sk_mse, mse.data))


class TestR2Score(unittest.TestCase):

    def test_forward1(self):
        y = np.array([1001.2, 211.3, 4500.1, 200.0])
        t = np.array([1231., 199.2, 4328., 176.])

        sk_r2 = metrics.r2_score(t, y)
        r2 = functions.evaluation.r2_score(y, t)

        self.assertEqual(sk_r2, r2.data)

    def test_forward2(self):
        y = np.array([[1001.2, 211.3, 4500.1, 200.0], [1421.2, 231.3, 40.1, 140.0]])
        t = np.array([[1231., 199.2, 4328., 176.], [1312., 220.2, 43.4, 116.]])

        sk_r2 = metrics.r2_score(t, y, multioutput="raw_values")
        r2 = functions.evaluation.r2_score(y, t, multi_output="raw_values")

        self.assertTrue(array_close(sk_r2, r2.data))

    def test_forward3(self):
        y = np.array([[1001.2, 211.3, 4500.1, 200.0], [1421.2, 231.3, 40.1, 140.0]])
        t = np.array([[1231., 199.2, 4328., 176.], [1312., 220.2, 43.4, 116.]])

        sk_r2 = metrics.r2_score(t, y)
        r2 = functions.evaluation.r2_score(y, t)

        self.assertTrue(array_close(sk_r2, r2.data))


class TestRMSE(unittest.TestCase):

    def test_forward1(self):
        y = np.array([1001.2, 211.3, 4500.1, 200.0])
        t = np.array([1231., 199.2, 4328., 176.])

        sk_rmse = metrics.mean_squared_error(t, y, squared=False)
        rmse = functions.evaluation.root_mean_squared_error(y, t)

        self.assertEqual(sk_rmse, rmse.data)

    def test_forward2(self):
        y = np.array([[1001.2, 211.3, 4500.1, 200.0], [1421.2, 231.3, 40.1, 140.0]])
        t = np.array([[1231., 199.2, 4328., 176.], [1312., 220.2, 43.4, 116.]])

        sk_rmse = metrics.mean_squared_error(t, y, multioutput="raw_values", squared=False)
        rmse = functions.evaluation.root_mean_squared_error(y, t, multi_output="raw_values")

        self.assertTrue(array_close(sk_rmse, rmse.data))

    def test_forward3(self):
        y = np.array([[1001.2, 211.3, 4500.1, 200.0], [1421.2, 231.3, 40.1, 140.0]])
        t = np.array([[1231., 199.2, 4328., 176.], [1312., 220.2, 43.4, 116.]])

        sk_rmse = metrics.mean_squared_error(t, y, squared=False)
        rmse = functions.evaluation.root_mean_squared_error(y, t)

        self.assertTrue(array_close(sk_rmse, rmse.data))


class TestRAE(unittest.TestCase):

    def test_forward1(self):
        y = np.array([1001.2, 211.3, 4500.1, 200.0])
        t = np.array([1231., 199.2, 4328., 176.])

        expected = np.absolute(y - t).sum() / np.absolute(t - t.mean()).sum()
        rae = functions.evaluation.relative_absolute_error(y, t)

        self.assertEqual(expected, rae.data)

    def test_forward2(self):
        y = np.array([[1001.2, 211.3, 4500.1, 200.0], [1421.2, 231.3, 40.1, 140.0]])
        t = np.array([[1231., 199.2, 4328., 176.], [1312., 220.2, 43.4, 116.]])

        expected = np.absolute(y - t).sum(axis=0) / np.absolute(t - t.mean(axis=0)).sum(axis=0)
        rae = functions.evaluation.relative_absolute_error(y, t, multi_output="raw_values")

        self.assertTrue(array_close(expected, rae.data))

    def test_forward3(self):
        y = np.array([[1001.2, 211.3, 4500.1, 200.0], [1421.2, 231.3, 40.1, 140.0]])
        t = np.array([[1231., 199.2, 4328., 176.], [1312., 220.2, 43.4, 116.]])

        expected = (np.absolute(y - t).sum(axis=0) / np.absolute(t - t.mean(axis=0)).sum(axis=0)).mean()
        rae = functions.evaluation.relative_absolute_error(y, t)

        self.assertTrue(array_close(expected, rae.data))


class TestRSE(unittest.TestCase):

    def test_forward1(self):
        y = np.array([1001.2, 211.3, 4500.1, 200.0])
        t = np.array([1231., 199.2, 4328., 176.])

        expected = np.square(y - t).sum() / np.square(t - t.mean()).sum()
        rse = functions.evaluation.relative_squared_error(y, t)

        self.assertEqual(expected, rse.data)

    def test_forward2(self):
        y = np.array([[1001.2, 211.3, 4500.1, 200.0], [1421.2, 231.3, 40.1, 140.0]])
        t = np.array([[1231., 199.2, 4328., 176.], [1312., 220.2, 43.4, 116.]])

        expected = np.square(y - t).sum(axis=0) / np.square(t - t.mean(axis=0)).sum(axis=0)
        rse = functions.evaluation.relative_squared_error(y, t, multi_output="raw_values")

        self.assertTrue(array_close(expected, rse.data))

    def test_forward3(self):
        y = np.array([[1001.2, 211.3, 4500.1, 200.0], [1421.2, 231.3, 40.1, 140.0]])
        t = np.array([[1231., 199.2, 4328., 176.], [1312., 220.2, 43.4, 116.]])

        expected = (np.square(y - t).sum(axis=0) / np.square(t - t.mean(axis=0)).sum(axis=0)).mean()
        rse = functions.evaluation.relative_squared_error(y, t)

        self.assertTrue(array_close(expected, rse.data))



import marquetry
from marquetry import configuration
from marquetry import cuda_backend
from marquetry import Function


class Dropout(Function):
    """Apply dropout to the input tensor.

        Dropout is a regularization technique used during training to prevent overfitting.
        It works by randomly deactivating neurons in the input tensor during each forward pass.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, dropout_rate):
        self.dropout_rate = dropout_rate

        self.mask = None

    def forward(self, x):
        if marquetry.configuration.config.train:
            xp = cuda_backend.get_array_module(x)
            mask = xp.random.rand(*x.shape) > self.dropout_rate
            self.mask = mask
            scale = xp.array(1.0 - self.dropout_rate).astype(x.dtype)
            y = x * mask / scale
        else:
            y = x

        self.retain_inputs(())

        return y

    def backward(self, x, grad_y):
        if configuration.config.train:
            grad_x = grad_y[0] * self.mask
        else:
            raise Exception("You execute non-train mode so you can't do backward.")

        return grad_x


def dropout(x, dropout_rate=0.5):
    """
        Apply dropout to the input tensor.

        Dropout is a regularization technique used during training to prevent overfitting. It works by randomly
        deactivating neurons in the input tensor during each forward pass.

        Args:
            x (:class:marquetry.Container or :class:numpy.ndarray or :class:cupy.ndarray):
                The input tensor.
            dropout_rate (float): The neurons to deactivate during each forward pass.
                It should be a float between 0 and 1, representing the probability of deactivation.
                Default is 0.5.

        Returns:
            marquetry.Container: The result of applying dropout to the input tensor.

        Note:
            During training (when `marquetry.configuration.config.train` is `True`),
                this function deactivates neurons in `x` randomly based on the specified `dropout_rate`.
                It also scales the remaining activations to maintain the expected value of the output.
            During inference (when `marquetry.configuration.config.train` is `False`),
                this function returns `x` unchanged.

    """

    return Dropout(dropout_rate)(x)



from marquetry import Model


class Sequential(Model):
    """A sequential neural network model for deep learning.

        The Sequential class allows you to define a neural network model by stacking multiple layers
        sequentially. This class is commonly used for building deep neural networks where layers are
        applied in a sequential order.

        Args:
            *layers_object: Variable-length argument list of layer objects. These layers will be stacked
                in the order they are passed.

        Architecture:
            Input Layer:
                The input data.
            Sequentially Stacked Layers:
                Multiple layers, such as convolutional layers, fully connected layers,
                and activation layers, stacked in the order they are defined.
            Output Layer:
                The final layer's output.

        Examples:
            >>> model = Sequential(
            >>>     marquetry.layers.Convolution2D(32, (3, 3)),
            >>>     marquetry.functions.relu,
            >>>     marquetry.layers.MaxPooling2D((2, 2)),
            >>>     marquetry.layers.Linear(128),
            >>>     marquetry.layers.Softmax(10))
            >>> y = model(data)

    """

    def __init__(self, *layers_object):
        super().__init__()
        self.layers = []

        if len(layers_object) == 1:
            if isinstance(layers_object[0], (tuple, list)):
                layers_object = tuple(layers_object[0])

        for i, layer in enumerate(layers_object):
            setattr(self, "l" + str(i), layer)
            self.layers.append(layer)

    def forward(self, x):
        for layer in self.layers:
            x = layer(x)

        return x



import numpy as np

import marquetry.cuda_backend as cuda_backend
import marquetry.utils as utils
from marquetry import functions
from marquetry import Layer
from marquetry import Parameter


class Convolution2D(Layer):
    """2D convolutional layer.

        The `Convolution2D` layer is commonly used in convolutional neural networks (CNNs) for tasks such as image
        recognition and generation. It performs 2D convolution operations on input data.

        Args:
            out_channels (int): The number of output channels (i.e., the number of filters or kernels).
            kernel_size (int or tuple of int): The size of the convolutional kernel.
            stride (int, optional): The stride of the convolution operation.
                Default is 1.
            pad (int): The amount of zero-padding to apply around the input data.
                Default is 0.
            nobias (bool): If True, no bias term is added to the convolution.
                Default is False.
            dtype (numpy.dtype or string): The data type used for weights and biases.
                Default is np.float32.
            in_channels (int or None): The number of input channels.

        Note:
            kernel_size:
                If an int is provided, it's treated as a square kernel size (e.g., 3 for a 3x3 kernel).
                If a tuple of int is provided, it specifies the height and width
                of the kernel (e.g., (3, 3) for a 3x3 kernel).
            in_channels:
                This parameter is automatically determined from the input data shape
                and does not need to be specified except a special use case.

        Attributes:
            w (marquetry.Parameter): The weight parameter.
            b (marquetry.Parameter): The bias parameter (if used).

        Examples:
            >>> x = np.arange(0, 60).reshape((1, 3, 4, 5))
            >>> x.shape
            (1, 3, 4, 5)
            >>> conv2d = Convolution2D(64, 3)
            >>> y = conv2d(x)
            >>> y.shape
            (1, 64, 2, 3)

    """
    def __init__(self, out_channels, kernel_size, stride=1, pad=0, nobias=False,
                 dtype=np.float32, in_channels=None):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.pad = pad
        self.dtype = dtype

        if nobias:
            self.b = None
        else:
            self.b = Parameter(np.zeros(out_channels, dtype=dtype), name="b")

        self.w = Parameter(None, name="w")

    def _init_w(self, xp=np):
        channels, output_channels = self.in_channels, self.out_channels
        kernel_height, kernel_width = utils.pair(self.kernel_size)

        scale = xp.sqrt(1 / (channels * kernel_height * kernel_width))
        w_data = xp.random.randn(output_channels, channels, kernel_height, kernel_width).astype(self.dtype) * scale

        self.w.data = w_data

        if self.b is not None and xp is not np:
            self.b.to_gpu()

    def forward(self, x):
        if self.w.data is None:
            self.in_channels = x.shape[1]
            xp = cuda_backend.get_array_module(x)
            self._init_w(xp=xp)

        y = functions.convolution_2d(x, self.w, self.b, self.stride, self.pad)

        return y



from marquetry import cuda_backend
from marquetry import Function


class Exp(Function):
    """Calculate the exponential of an input.

        This class computes the exponential of the input.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.exp(x)

        self.retain_inputs(())
        self.retain_outputs((0,))
        return y

    def backward(self, x, grad_y):
        y = self.output_data[0]
        grad_x = grad_y[0] * y

        return grad_x


def exp(x):
    """Calculate the exponential of an input.

        This function computes the exponential of the input.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input values.

        Returns:
            :class:`marquetry.Container`: The exponential of the input.

        Examples:
            >>> x = np.array([1, 2, 3])
            >>> exp(x)
            container([ 2.71828183  7.3890561  20.08553692])
    """

    return Exp()(x)


class Log(Function):
    """Calculate the natural logarithm of an input.

        This class computes the natural logarithm of the input.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.log(x)

        return y

    def backward(self, x, grad_y):
        grad_x = grad_y[0] / x[0]

        return grad_x


def log(x):
    """Calculate the natural logarithm of an input.

        This function computes the natural logarithm of the input.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input values.

        Returns:
            :class:`marquetry.Container`: The natural logarithm of the input.

        Examples:
            >>> x = np.array([1, 2, 3])
            >>> log(x)
            container([0.         0.69314718 1.09861229])

    """

    return Log()(x)


class Log2(Function):
    """Calculate the base-2 logarithm of an input.

        This class computes the base-2 logarithm of the input.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.log2(x)

        return y

    def backward(self, x, grad_y):
        grad_x = grad_y[0] / (x[0] * log(2))

        return grad_x


def log2(x):
    """Calculate the base-2 logarithm of an input.

        This function computes the base-2 logarithm of the input.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input values.

        Returns:
            :class:`marquetry.Container`: The base-2 logarithm of the input.

        Examples:
            >>> x = np.array([1, 2, 3])
            >>> log2(x)
            container([0.        1.        1.5849625])
    """
    return Log2()(x)


class Log10(Function):
    """Calculate the base-10 logarithm of an input.

        This function computes the base-10 logarithm of the input.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)
        y = xp.log10(x)

        return y

    def backward(self, x, grad_y):
        grad_x = grad_y[0] / (x[0] * log(10))

        return grad_x


def log10(x):
    """Calculate the base-10 logarithm of an input.

        This function computes the base-10 logarithm of the input.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The input values.

        Returns:
            :class:`marquetry.Container`: The base-10 logarithm of the input.

        Examples:
            >>> x = np.array([1, 2, 3])
            >>> log10(x)
            container([0.         0.30103    0.47712125])
    """

    return Log10()(x)



import marquetry


def information_gain(target, target_left, target_right, criterion=None, target_type="classification"):
    """Compute the information gain from splitting a dataset into two subsets.

        Information gain measures how much the impurity decreases from before splitting
        to after splitting a dataset into two subsets.
        It is commonly used in decision tree algorithms to evaluate the quality of a split.

        Args:
            target (:class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The target container before splitting.
            target_left (:class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The target container of the left subset after splitting.
            target_right (:class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The target container of the right subset after splitting.
            criterion (str or None): The impurity criterion to use.
                For classification, options are "gini" or "entropy."
                For regression, options are "rss" (Residual Sum of Squares) or "mae" (Mean Absolute Error).
                Default is None, which selects the appropriate criterion based on the target_type.
            target_type (str): The type of the target container.
                Options are "classification" or "regression."
                Default is "classification."

        Returns:
            float: The computed information gain.

    """
    impurity_target = marquetry.utils.impurity_criterion(target, criterion=criterion, target_type=target_type)
    impurity_left = marquetry.utils.impurity_criterion(target_left, criterion=criterion, target_type=target_type)
    impurity_right = marquetry.utils.impurity_criterion(target_right, criterion=criterion, target_type=target_type)

    split_mean_impurity = (float(len(target_left) / len(target)) * impurity_left +
                           float(len(target_right) / len(target) * impurity_right))
    info_gain = impurity_target - split_mean_impurity

    return info_gain



from marquetry import cuda_backend
from marquetry import Function


class Precision(Function):
    """Compute precision based on true labels and predicted values using a specified threshold.

        This class defines a function that computes precision based on the true binary labels and predicted values.
        The precision indicates the accuracy of the prediction, more easily, this evaluator indicates
        how many correct the positive prediction compare the all positive prediction.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, threshold, dtype="float64"):
        if not 0. <= threshold <= 1.:
            raise ValueError("threshold should be in (0.0, 1.0), but got {}".format(threshold))
        self.threshold = threshold
        self.dtype = dtype

    def forward(self, y, t):
        xp = cuda_backend.get_array_module(y)

        assert len(xp.unique(t)) <= 2

        pred = xp.asarray((y >= self.threshold), dtype="f").reshape(t.shape)

        pred_positive_num = pred.sum()
        true_positive_num = pred[t == 1].sum()

        self.retain_inputs(())
        if pred_positive_num == 0:
            return xp.asarray(0.0, dtype=self.dtype)
        else:
            return xp.asarray(true_positive_num / pred_positive_num, dtype=self.dtype)


def precision(y, t, threshold=0.7, dtype="float64"):
    """Compute precision based on true labels and predicted values using a specified threshold.

        This function defines that computes precision based on the true binary labels and predicted values.
        The precision indicates the accuracy of the prediction, more easily, this evaluator indicates
        how many correct the positive prediction compare the all positive prediction.

        Precision is calculated as:
            :math:`Precision = TP / TP + FP`
            (TP -> TruePositive, FP -> FalsePositive)

        Args:
            y (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted values.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true binary labels.
            threshold (float): The threshold value used to determine binary predictions.
                Defaults to 0.7.
            dtype (str or numpy.dtype):
                The return value's dtype, Default is "float64" that means 64-bit float value.

        Returns:
            :class:`marquetry.Container`: The precision based on the true binary labels
                and predicted values using the specified threshold.
    """

    return Precision(threshold, dtype)(y, t)


class MultiPrecision(Function):
    """Calculate precision for a specific target class based on true labels and predicted values.

        This class defines a function that calculates precision for a specific target class based on
        the true labels and predicted values.
        Precision measures the accuracy of positive predictions for the specified target class,
        indicating how many positive predictions are correct compare the all positive prediction.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """

    def __init__(self, target_class, dtype="float64"):
        self.target_class = target_class
        self.dtype = dtype

    def forward(self, y, t):
        xp = cuda_backend.get_array_module(y)

        assert len(xp.unique(t)) > 2

        if y.size != t.size:
            pred = xp.argmax(y, axis=1).reshape(t.shape)
        else:
            pred = y

        pred_positive = xp.asarray(pred == self.target_class, dtype="f")
        pred_true_map = xp.asarray(pred == t, dtype="f")

        pred_true_positive = (pred_positive * pred_true_map).sum()
        pred_positive_num = pred_positive.sum()

        self.retain_inputs(())
        if pred_positive_num == 0:
            return xp.asarray(0.0, dtype=self.dtype)
        else:
            return xp.asarray(pred_true_positive / pred_positive_num, dtype=self.dtype)


def multi_precision(y, t, target_class=1, dtype="float64"):
    """Calculate precision for a specific target class based on true labels and predicted values.


        This function defines that calculates precision for a specific target class based on
        the true labels and predicted values.
        Precision measures the accuracy of positive predictions for the specified target class, indicating how many
        positive predictions are correct compare the all positive prediction.

        Precision is calculated as:
            :math:`Precision = TP / TP + FP`
            (TP -> True Positives, FP -> False Positives)

        Args:
            y (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The predicted values.
            t (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                The true labels.
            target_class (int): The class index to calculate precision for.
                Defaults to 1.
            dtype (str or numpy.dtype):
                The return value's dtype, Default is "float64" that means 64-bit float value.

        Returns:
            :class:`marquetry.Container`:
                The precision for the specified target class based on the true labels and predicted values.
    """

    return MultiPrecision(target_class, dtype)(y, t)



import unittest

import numpy as np
import torch

import marquetry.functions as funcs
from marquetry import Container
from marquetry.utils import array_close, gradient_check


class TestMax(unittest.TestCase):

    def test_forward1(self):
        x = Container(np.random.rand(10))
        y = funcs.max(x)

        res = y.data
        expected = np.max(x.data)

        self.assertTrue(array_close(res, expected))

    def test_forward2(self):
        shape = (10, 20, 30)
        axis = 1

        x = Container(np.random.randn(*shape))
        y = funcs.max(x, axis=axis)

        res = y.data
        expected = np.max(x.data, axis=axis)

        self.assertTrue(array_close(res, expected))

    def test_forward3(self):
        shape = (10, 20, 30)
        axis = (0, 1)

        x = Container(np.random.randn(*shape))
        y = funcs.max(x, axis=axis)

        res = y.data
        expected = np.max(x.data, axis=axis)

        self.assertTrue(array_close(res, expected))

    def test_forward4(self):
        shape = (10, 20, 30)
        axis = (0, 1)

        x = Container(np.random.randn(*shape))
        y = funcs.max(x, axis=axis, keepdims=True)

        res = y.data
        expected = np.max(x.data, axis=axis, keepdims=True)

        self.assertTrue(array_close(res, expected))

    def test_backward1(self):
        x_data = np.random.randn(10)
        f = lambda x: funcs.max(x)

        self.assertTrue(gradient_check(f, x_data))

    def test_backward2(self):
        x_data = np.random.randn(10, 10)
        f = lambda x: funcs.max(x, axis=1)

        self.assertTrue(gradient_check(f, x_data))

    def test_backward3(self):
        x_data = np.random.randn(10, 20, 30)
        f = lambda x: funcs.max(x, axis=(1, 2))

        self.assertTrue(gradient_check(f, x_data))

    def test_backward4(self):
        x_data = np.random.randn(10, 20, 20)
        f = lambda x: funcs.max(x, axis=None)

        self.assertTrue(gradient_check(f, x_data))

    def test_backward5(self):
        x_data = np.random.randn(10, 20, 20)
        f = lambda x: funcs.max(x, axis=None, keepdims=True)

        self.assertTrue(gradient_check(f, x_data))


class TestMin(unittest.TestCase):

    def test_forward1(self):
        x = Container(np.random.rand(10))
        y = funcs.min(x)

        res = y.data
        expected = np.min(x.data)

        self.assertTrue(array_close(res, expected))

    def test_forward2(self):
        shape = (10, 20, 30)
        axis = 1

        x = Container(np.random.randn(*shape))
        y = funcs.min(x, axis=axis)

        res = y.data
        expected = np.min(x.data, axis=axis)

        self.assertTrue(array_close(res, expected))

    def test_forward3(self):
        shape = (10, 20, 30)
        axis = (0, 1)

        x = Container(np.random.randn(*shape))
        y = funcs.min(x, axis=axis)

        res = y.data
        expected = np.min(x.data, axis=axis)

        self.assertTrue(array_close(res, expected))

    def test_forward4(self):
        shape = (10, 20, 30)
        axis = (0, 1)

        x = Container(np.random.randn(*shape))
        y = funcs.min(x, axis=axis, keepdims=True)

        res = y.data
        expected = np.min(x.data, axis=axis, keepdims=True)

        self.assertTrue(array_close(res, expected))

    def test_backward1(self):
        x_data = np.random.randn(10)
        f = lambda x: funcs.min(x)

        self.assertTrue(gradient_check(f, x_data))

    def test_backward2(self):
        x_data = np.random.randn(10, 10)
        f = lambda x: funcs.min(x, axis=1)

        self.assertTrue(gradient_check(f, x_data))

    def test_backward3(self):
        x_data = np.random.randn(10, 20, 30)
        f = lambda x: funcs.min(x, axis=(1, 2))

        self.assertTrue(gradient_check(f, x_data))

    def test_backward4(self):
        x_data = np.random.randn(10, 20, 20)
        f = lambda x: funcs.min(x, axis=None)

        self.assertTrue(gradient_check(f, x_data))

    def test_backward5(self):
        x_data = np.random.randn(10, 20, 20)
        f = lambda x: funcs.min(x, axis=None, keepdims=True)

        self.assertTrue(gradient_check(f, x_data))


class TestClip(unittest.TestCase):

    def test_forward1(self):
        x = np.array([[0, .3, .8], [1.2, .3, .9], [0, .2, 1.5]])
        y = funcs.clip(x, 1e-5, 1.)

        res = y.data
        expected = np.array([[1e-5, .3, .8], [1.0, .3, .9], [1e-5, .2, 1.0]])

        self.assertTrue(array_close(res, expected))

    def test_forward2(self):
        x = np.array([[0, .3, .8], [1.2, .3, .9], [0, .2, 1.5]])
        y = funcs.clip(x, x_min=1e-5, x_max=1.)
        ty = torch.clip(torch.tensor(x), min=torch.tensor(1e-5), max=torch.tensor(1.))

        self.assertTrue(array_close(y.data, ty.data.numpy()))



from typing import Literal

import pandas as pd

from marquetry.preprocesses.encodes.label_encode import LabelEncode
from marquetry.preprocesses.encodes.one_hot_encode import OneHotEncode

from marquetry.preprocesses.imputation.miss_imputation import MissImputation

from marquetry.preprocesses.normalize.column_normalize import ColumnNormalize
from marquetry.preprocesses.normalize.column_standardize import ColumnStandardize


class ToEncodeData(object):
    """Data Encoding and Preprocessing Utility.

    ToEncodeData is a utility class that facilitates data encoding and preprocessing for machine learning tasks.
    It combines several preprocessing steps such as label encoding, one-hot encoding, imputation, and normalization.

    Args:
        target_column (str): The name of the target column in the dataset.
        category_columns (list): A list of column names to be treated as categorical variables.
        numeric_columns (list): A list of column names to be treated as numeric variables.
        name (str): The name of the ToEncodeData instance.
        is_train (bool): A boolean flag indicating whether this instance is used for training or inference.
        imputation_category_method (str): The imputation method for categorical columns.
            Choices: "mode" (default) or "zero".
        imputation_numeric_method (str, optional): The imputation method for numeric columns.
            Choices: "mean" (default), "mode", "median", or "zero".
        label_encode_columns (list or None): A list of column names to be label encoded.
            Default is None which means no columns to label encoding.
        one_hot_encode_columns (list or None): A list of column names to be one-hot encoded.
            Default is None which means one-hot encode all category columns.
        include_null (bool): If True, include null values in encoding and imputation.
            Default is False.
        normalize_method (str): The method for normalizing columns.
            Choices: "normalize" (default), "standardize", or None (no normalization).
        specify_labels (dict or None): A dictionary specifying label mappings for label encoding.
            Default is None.
        allow_unknown_category (bool): If True, allow encoding of unknown categories.
            Default is True.

    Examples:
        >>> import pandas as pd
        >>> from marquetry.preprocesses import ToEncodeData
        >>> data = pd.DataFrame({'Category': ['A', 'B', 'A', 'C'], 'Numeric': [10, 20, 30, 40]})
        >>> encoder = ToEncodeData(target_column='Category', category_columns=['Category'], numeric_columns=['Numeric'],
        >>>                        name='encoder', is_train=True, allow_unknown_category=False)
        >>> encoded_data, unique_values = encoder(data)
        >>> print(encoded_data)
            Numeric  Category_B  Category_C
        0 -1.161895           0           0
        1 -0.387298           1           0
        2  0.387298           0           0
        3  1.161895           0           1
        >>> print(unique_values)
        {'Category': 3, 'Numeric': 'N/A'}

    Note:
        ToEncodeData combines label encoding, one-hot encoding, imputation, and normalization in one utility class.
        You can specify various encoding and imputation methods as well as which columns to encode.
        The `remove_old_statistic` method is provided to clear old statistics if needed.

    See Also:
        - :class:`LabelEncode`: Class for label encoding categorical columns.
        - :class:`OneHotEncode`: Class for one-hot encoding categorical columns.
        - :class:`MissImputation`: Class for imputing missing values in data.
        - :class:`ColumnNormalize`: Class for normalizing columns.
        - :class:`ColumnStandardize`: Class for standardizing columns.
    """

    def __init__(self,
                 target_column,
                 category_columns,
                 numeric_columns,
                 name,
                 is_train: bool,
                 imputation_category_method="mode",
                 imputation_numeric_method="mean",
                 label_encode_columns=None,
                 one_hot_encode_columns=None,
                 include_null=False,
                 normalize_method: Literal["normalize", "standardize", None] = "normalize",
                 specify_labels: dict = None,
                 allow_unknown_category=True):

        self._name = name
        self._target_column = target_column
        self._category_columns = category_columns
        self._numeric_columns = numeric_columns
        self._imputation_category_method = imputation_category_method
        self._imputation_numeric_method = imputation_numeric_method

        self._label_encode_columns = label_encode_columns if label_encode_columns is not None else []
        self._one_hot_encode_columns = \
            one_hot_encode_columns if one_hot_encode_columns is not None else self._category_columns

        self._one_hot_encoder = OneHotEncode(
            self._category_columns, self._name, is_train, include_null, allow_unknown_category)

        unknown_handler: Literal["encode_specify_value", "raise_error"]
        if allow_unknown_category:
            unknown_handler = "encode_specify_value"
        else:
            unknown_handler = "raise_error"

        self._label_encoder = LabelEncode(self._category_columns, self._name, is_train, specify_label=specify_labels,
                                          include_null=include_null, treat_unknown=unknown_handler)

        if normalize_method == "normalize":
            self._norm_calculator = (
                ColumnNormalize(self._numeric_columns + self._category_columns, self._name, is_train))
        elif normalize_method == "standardize":
            self._norm_calculator = (
                ColumnStandardize(self._numeric_columns + self._category_columns, self._name, is_train))
        else:
            self._norm_calculator = None

        self._imputation_runner = MissImputation(self._category_columns, self._numeric_columns, self._name, is_train,
                                                 self._imputation_category_method, self._imputation_numeric_method)

    def __call__(self, data):
        data = self._imputation_runner(data)
        labeled_table = self._label_encoder(data)
        one_hot_table = self._one_hot_encoder(data)

        unique_dict = {}
        for column_name in data.keys():
            if column_name in self._category_columns:
                unique_num = int(labeled_table[column_name].max()) + 1
                unique_dict[column_name] = unique_num
            else:
                unique_dict[column_name] = "N/A"

        labeled_table = labeled_table[self._category_columns]
        data = pd.concat((data.drop(self._category_columns, axis=1), labeled_table), axis=1)

        if self._norm_calculator is not None:
            data = self._norm_calculator(data)

        one_hot_columns = \
            [column for column in one_hot_table.columns if column.split("_")[0] in self._one_hot_encode_columns]
        one_hot_table = one_hot_table[one_hot_columns]

        output_table = pd.concat(
            (data.drop(self._one_hot_encode_columns, axis=1), one_hot_table), axis=1)

        return output_table, unique_dict

    def remove_old_statistic(self):
        """Remove old statistics used for imputation, normalization, and encoding."""
        if self._norm_calculator is not None:
            self._norm_calculator.remove_old_statistic()

        self._imputation_runner.remove_old_statistic()
        self._label_encoder.remove_old_statistic()
        self._one_hot_encoder.remove_old_statistic()



from marquetry import Function
from marquetry import utils
from marquetry import functions


class LogSoftmax(Function):
    """Log Softmax.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def __init__(self, axis):
        self.axis = axis

    def forward(self, x):
        log_z = utils.log_sum_exp(x, self.axis)
        y = x - log_z

        self.retain_inputs(())
        self.retain_outputs((0,))
        return y

    def backward(self, x, grad_y):
        y = self.output_data[0]
        grad_y = grad_y[0]

        grad_x = grad_y - functions.exp(y) * grad_y.sum(axis=self.axis, keepdims=True)

        return grad_x


def log_softmax(x, axis=1):
    """Log Softmax.

        :math:`f(x) = log{exp(x) / Σexp(x)}`
            - The ``log`` is a natural logarithm.

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                Input container that is float array.
            axis (int): The sum calc axis.

        Returns:
            marquetry.Container: Output container. A float array.

        Examples:

            >>> x = np.array([[-1, 0], [2, -3], [-2, 1]], 'f')
            >>> x
            array([[-1.,  0.],
                    [ 2., -3.],
                    [-2.,  1.]], dtype=float32)
            >>> log_softmax(x, axis=1)
            container([[-1.3132616  -0.31326166]
                       [-0.0067153  -5.0067153 ]
                       [-3.0485873  -0.04858732]])

    """
    return LogSoftmax(axis)(x)



import numpy as np
import pandas as pd

from marquetry import Preprocess


class OneHotEncode(Preprocess):
    """One-Hot Encoding for Categorical Data.

        OneHotEncode is a preprocessing step used to one-hot encode categorical columns in a Pandas DataFrame.
        This transformation converts categorical columns into a set of binary columns,
        where each unique category is represented as a separate binary column.

        Args:
            category_column (list): A list of column names to be one-hot encoded.
            name (str): The name of the OneHotEncode instance.
            is_train (bool): A boolean flag indicating whether this OneHotEncode instance is for training or inference.
            include_null (bool): Whether to include null or NaN values when calculating one-hot encoding.
                Defaults to False.
            allow_unknown_value (bool): Whether to allow unknown categories during one-hot encoding.
                If set to True, an additional binary column will be created for unknown categories.
                Defaults to False.

        Examples:
            >>> import pandas as pd
            >>> from marquetry.preprocesses import OneHotEncode
            >>> data = pd.DataFrame({'Category': ['A', 'B', 'A', 'C']})
            >>> one_hot_encoder = OneHotEncode(category_column=['Category'], name='one_hot_encoder', is_train=True)
            >>> encoded_data = one_hot_encoder(data)
            >>> print(encoded_data)
               Category_B  Category_C
            0           0           0
            1           1           0
            2           0           0
            3           0           1

        Note:
            OneHotEncode converts categorical columns into binary columns (0 and 1).
            And, You can control the inclusion of null or NaN values using the `include_null` parameter and,
            the `allow_unknown_value` parameter allows you to create an additional binary column for unknown categories.

            To avoid multicolinearity problem, remove the one binary category column.
    """

    _label = "pre_ohe"

    def __init__(self, category_column: list, name: str, is_train: bool, include_null=False, allow_unknown_value=False):
        super().__init__(name, is_train)
        self._category_column = category_column
        self._include_null = include_null
        self._unknown_value = allow_unknown_value

    def process(self, data: pd.DataFrame):
        """Process the input DataFrame by one-hot encoding specified columns.

            Args:
                data (:class:`pandas.DataFrame`): The input DataFrame to be one-hot encoded.

            Returns:
                pd.DataFrame: The one-hot encoded DataFrame.

            Caution:
                Generally, process will be called by ``Marquetry`` core.
                Please call like the ``EXAMPLES``.

        """

        if len(self._category_column) == 0:
            return data

        type_change_dict = {key: str for key in self._category_column}
        data = data.astype(type_change_dict).replace("nan", np.nan)

        batch_size = len(data)

        if self._statistic_data is None:
            self._calc_statistic(data)

        self._validate_values(data)

        replaced_data = data.replace(self._statistic_data)
        replaced_data = replaced_data[self._category_column]

        one_hot_data = None
        for column in self._category_column:
            tmp_dict = self._statistic_data[column]
            unique_nums = len(tmp_dict)
            column_names = [column + "_" + str(key) for i, key in enumerate(tmp_dict.keys()) if i != 0]

            # unknown value replacing process
            unknown_replace_dict = {}
            if self._unknown_value:
                unique_set = set(data[column])
                statistic_set = set(pd.Series(tmp_dict).keys())
                unknown_values = list(unique_set - statistic_set)
                next_num = unique_nums
                unknown_replace_dict = {key: next_num for key in unknown_values}
                column_names.append(column + "_" + "unknown")
                unique_nums += 1

            # create one-hot table
            one_hot_array = np.zeros((batch_size, unique_nums))
            one_hot_array[
                np.arange(batch_size), list(replaced_data[column].replace(unknown_replace_dict).astype(np.int32))] = 1.
            one_hot_df = pd.DataFrame(one_hot_array[:, 1:].astype(np.int32), columns=column_names)

            if one_hot_data is None:
                one_hot_data = one_hot_df
            else:
                one_hot_data = pd.concat((one_hot_data, one_hot_df), axis=1)

        return one_hot_data

    def _validate_values(self, data):
        exist_statistic_columns = set([key for key, value in self._statistic_data.items() if value != {}])
        input_target_columns = set(self._category_column)

        if exist_statistic_columns != input_target_columns:
            raise ValueError("saved statistic data's target columns is {} but you input {}. "
                             .format(exist_statistic_columns, input_target_columns) + self._msg)

        for column in self._category_column:
            unique_set = set(data[column])
            if not self._include_null:
                unique_set = set(unique_value for unique_value in unique_set if not pd.isna(unique_value))

            statistic_set = set(pd.Series(self._statistic_data[column]).keys())

            if unique_set != statistic_set and not self._unknown_value:
                raise ValueError("statistic data doesn't have {} category in '{}' but the input has it. "
                                 .format(",".join(sorted(list(unique_set - statistic_set))), column) + self._msg)

        return

    def _calc_statistic(self, data: pd.DataFrame):
        one_hot_dict = {}
        for column in list(data.columns):
            if column not in self._category_column:
                one_hot_dict[column] = {}
                continue

            tmp_series = data[column]
            unique_set = list(set(tmp_series))

            if not self._include_null:
                unique_set = [unique_value for unique_value in unique_set if not pd.isna(unique_value)]

            class_nums = list(range(len(unique_set)))

            tmp_dict = dict(zip(sorted(unique_set), class_nums))
            one_hot_dict[column] = tmp_dict

        self._save_statistic(one_hot_dict)
        self._statistic_data = one_hot_dict



import marquetry.cuda_backend as cuda_backend
from marquetry import functions
from marquetry import Layer
from marquetry import Parameter


class BatchNormalization(Layer):
    """Batch normalization layer.

        Batch normalization is a technique used in deep neural networks to stabilize and accelerate training.
        This layer normalizes the input data by using normalization and adjustable parameters (gamma and beta).
        It helps prevent issues like vanishing gradients, reduces co-variate shift,
        and allows for faster convergence.

        Args:
            decay (float): The weighting factor for the moving averages of mean and variance.
                A smaller value will make the moving averages change more slowly.
                Default is 0.9.

        Attributes:
            gamma (marquetry.Parameter): The gamma parameter used for scaling.
            beta (marquetry.Parameter): The beta parameter used for shifting.
            avg_mean (marquetry.Parameter): The moving average of the mean.
            avg_var (marquetry.Parameter): The moving average of the variance.

        Examples:
            >>> x = np.random.rand(4, 5)
            >>> x.mean(axis=0)
            array([0.34748032, 0.55353551, 0.46282959, 0.4974482 , 0.60365538])
            >>> x.std(axis=0)
            array([0.25182043, 0.25240281, 0.29050406, 0.37605534, 0.30533898])
            >>> batch_norm = BatchNormalization()
            >>> y = batch_norm(x)
            >>> y.data.mean(axis=0)
            array([0.00000000e+00, 0.00000000e+00,  0.00000000e+00,  0.00000000e+00,  0.00000000e+00])
            (Depending on the computing environment, the mean is not 0.0 but the gap is very small
            (generally under 1e-15).)
            >>> y.data.std(axis=0)
            array([1., 1., 1., 1., 1.])


    """

    def __init__(self, decay=0.9):
        super().__init__()
        self.avg_mean = Parameter(None, name="avg_mean")
        self.avg_var = Parameter(None, name="avg_var")
        self.gamma = Parameter(None, name="gamma")
        self.beta = Parameter(None, name="beta")

        self.decay = decay

    def __call__(self, x):
        xp = cuda_backend.get_array_module(x)
        if self.avg_mean.data is None:
            input_shape = x.shape[1]
            if self.avg_mean.data is None:
                self.avg_mean.data = xp.zeros(input_shape, dtype=x.dtype)
            if self.avg_var.data is None:
                self.avg_var.data = xp.ones(input_shape, dtype=x.dtype)
            if self.gamma.data is None:
                self.gamma.data = xp.ones(input_shape, dtype=x.dtype)
            if self.beta.data is None:
                self.beta.data = xp.zeros(input_shape, dtype=x.dtype)

        return functions.batch_normalization(x, self.gamma, self.beta, self.avg_mean.data, self.avg_var.data, self.decay)



from marquetry.container import as_array
from marquetry.container import as_container
from marquetry.container import array
from marquetry.container import Parameter
from marquetry.container import random
from marquetry.container import random_gen
from marquetry.container import random_int
from marquetry.container import Container

from marquetry.function import Function

from marquetry.machine_learning import MachineLearning

from marquetry.configuration import config
from marquetry.configuration import using_config
from marquetry.function import no_backprop_mode
from marquetry.function import test_mode

from marquetry.layer import Layer
from marquetry.model import Model
from marquetry.optimizer import WeightDecay
from marquetry.optimizer import ClipGrad
from marquetry.preprocess import Preprocess

import marquetry.dataloaders
import marquetry.datasets
import marquetry.functions
import marquetry.layers
import marquetry.ml
import marquetry.models
import marquetry.optimizers
import marquetry.preprocesses
import marquetry.cuda_backend
import marquetry.transformers
import marquetry.utils
import marquetry.configuration



from marquetry import cuda_backend


def log_sum_exp(x, axis=1):
    """Compute the log-sum-exp of an array along a specified axis.

        This function computes the log-sum-exp of an input array along the specified axis.
        It first subtracts the maximum value along the axis from the input array to prevent overflow,
        exponentiates the result, computes the sum along the axis, and finally takes the logarithm of the sum.

        logsumexp is used for preventing over/underflow the exponential generally.

        Args:
            x (:class:`numpy.ndarray`, or :class:`cupy.ndarray`):
                The input array.
            axis (int): The axis along which the log-sum-exp should be computed.
                Default is axis 1.

        Returns:
            :class:`numpy.ndarray` or :class:`cupy.ndarray`:
                The log-sum-exp of the input array along the specified axis.
    """

    xp = cuda_backend.get_array_module(x)
    x_max = x.max(axis=axis, keepdims=True)
    x_norm = x - x_max
    xp.exp(x_norm, out=x_norm)
    sum_exp = x_norm.sum(axis=axis, keepdims=True)
    xp.log(sum_exp, out=sum_exp)

    y = x_max + sum_exp

    return y



from marquetry import cuda_backend
from marquetry import Function


class Softmax(Function):
    """Softmax activation function.

        Note:
            Generally, you don't need to execute ``forward`` and ``backward`` method manually.
            You should use only ``__call__`` method.
    """
    def __init__(self, axis):
        self.axis = axis

    def forward(self, x):
        xp = cuda_backend.get_array_module(x)

        y = x - x.max(axis=self.axis, keepdims=True)
        y = xp.exp(y)
        y /= y.sum(axis=self.axis, keepdims=True)

        self.retain_inputs(())
        self.retain_outputs((0,))
        return y

    def backward(self, x, grad_y):
        y = self.output_data[0]
        grad_x = y * grad_y[0]
        sum_grad_x = grad_x.sum(axis=self.axis, keepdims=True)
        grad_x -= y * sum_grad_x

        return grad_x


def softmax(x, axis=1):
    """Softmax function.

        :math:`f(x) = exp(x) / Σexp(x)`

        Args:
            x (:class:`marquetry.Container` or :class:`numpy.ndarray` or :class:`cupy.ndarray`):
                Input container that is float array.
            axis (int): The softmax sum axis

        Returns:
            marquetry.Container: Output container. A float array.

        Examples:

            >>> x = np.array([[-1, 0], [2, -3], [-2, 1]], 'f')
            >>> x
            array([[-1.,  0.],
                   [ 2., -3.],
                   [-2.,  1.]], dtype=float32)
            >>> softmax(x, axis=1)
            container([[0.26894143 0.7310586 ]
                       [0.9933072  0.00669285]
                       [0.04742587 0.95257413]])

    """

    return Softmax(axis)(x)



import unittest

import numpy as np

import marquetry.functions as funcs
from marquetry import Container
from marquetry.utils import array_equal, gradient_check


class TestConcat(unittest.TestCase):

    def test_forward1(self):
        x0 = Container(np.array([[1, 2], [3, 4], [5, 6]]))
        x1 = np.array([[4], [4], [4]])

        y = funcs.concat((x0, x1), axis=1)

        self.assertEqual(y.shape, (3, 3))
        self.assertTrue(array_equal(y[:, 2].data.reshape((-1, 1)), x1))

    def test_forward2(self):
        x0 = np.array([[1, 2, 3], [4, 5, 6]])
        x1 = np.array([[2, 4, 6]])

        y = funcs.concat(x0, x1)

        self.assertEqual(y.shape, (3, 3))
        self.assertTrue(array_equal(y[2].reshape(1, -1), x1))

    def test_backward1(self):
        x0 = np.random.randn(2, 3)
        x1 = np.random.randn(1, 3)

        f = lambda x: funcs.concat((x, x1), axis=0)

        self.assertTrue(gradient_check(f, x0))

    def test_backward2(self):
        x0 = np.random.randn(3, 2)
        x1 = np.random.randn(3, 2)

        f = lambda x: funcs.concat((x0, x), axis=1)

        self.assertTrue(gradient_check(f, x1))


class TestSplit(unittest.TestCase):

    def test_forward1(self):
        x = np.array([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]])
        y0, y1 = funcs.split(x, 2, axis=1)

        self.assertTrue(array_equal(x[:, :2], y0.data))
        self.assertTrue(array_equal(x[:, 2:], y1.data))

    def test_forward2(self):
        x = np.array([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6], [3, 4, 5, 6, 7]])
        y0, y1, y2 = funcs.split(x, (1, 3), axis=1)

        self.assertTrue(array_equal(x[:, :1], y0.data))
        self.assertTrue(array_equal(x[:, 1:3], y1.data))
        self.assertTrue(array_equal(x[:, 3:], y2.data))

    def test_backward1(self):
        x = np.random.randn(2, 10)

        def func(x):
            y0, y1 = funcs.split(x, indices=5, axis=1)
            return y0 + y1

        self.assertTrue(gradient_check(func, x))

    def test_backward2(self):
        x = np.random.randn(2, 10)

        def func(x):
            y0, y1 = funcs.split(x, indices=1, axis=0)
            return y0 * y1

        self.assertTrue(gradient_check(func, x))

    def test_backward3(self):
        x = np.random.randn(3, 12)

        def func(x):
            y0, y1, y2 = funcs.split(x, indices=(4, 8), axis=1)
            return y0 - y1 - y2

        self.assertTrue(gradient_check(func, x))


class TestSqueeze(unittest.TestCase):

    def test_forward1(self):
        x = np.random.randn(3, 1, 4)
        y = funcs.squeeze(x, axis=1)

        self.assertEqual(y.shape, (3, 4))

    def test_forward2(self):
        x = np.random.randn(3, 2, 4)

        with self.assertRaises(ValueError):
            funcs.squeeze(x, axis=1)

    def test_backward1(self):
        x = np.random.randn(3, 1, 5)

        self.assertTrue(gradient_check(funcs.squeeze, x, axis=1))

    def test_backward2(self):
        x = np.random.randn(3, 1, 5, 1, 2, 1)
        axis = (1, 3, 5)

        f = lambda x: funcs.squeeze(x, axis=axis)

        self.assertTrue(gradient_check(f, x))

    def test_shape_check(self):
        x = Container(np.random.randn(3, 1, 5))
        y = funcs.squeeze(x, axis=1)
        y.backward()

        self.assertTrue(x.grad.shape, x.shape)


class TestUnSqueeze(unittest.TestCase):

    def test_forward1(self):
        x = np.random.randn(3, 4)
        y = funcs.unsqueeze(x, axis=1)

        self.assertEqual(y.shape, (3, 1, 4))

    def test_forward2(self):
        x = Container(np.random.randn(3, 4))
        y = x.unsqueeze(axis=0)

        self.assertEqual(y.shape, (1, 3, 4))

    def test_backward1(self):
        x = np.random.randn(3, 4)

        self.assertTrue(gradient_check(funcs.unsqueeze, x, axis=1))

    def test_backward2(self):
        x = np.random.randn(3, 5, 2)
        axis = (1, 3, 5)

        f = lambda x: funcs.unsqueeze(x, axis=axis)

        self.assertTrue(gradient_check(f, x))

    def test_shape_check(self):
        x = Container(np.random.randn(3, 4))
        y = x.unsqueeze(axis=2)
        y.backward()

        self.assertEqual(x.grad.shape, x.shape)
